# -*- encoding: utf-8 -*-
"""
GObj :class:`GHttpCliSrv`
=========================

Http Channel.

.. autoclass:: GHttpCliSrv
    :members:

"""
from collections import deque
import traceback

from ginsfsm.c_timer import GTimer
from ginsfsm.gobj import GObj
from ginsfsm.gconfig import GConfig
from ginsfsm.protocols.http.common.parser import HTTPRequestParser
from ginsfsm.protocols.http.common.utilities import InternalServerError
from ginsfsm.protocols.http.common.response import (
    HttpResponse,
    HttpErrorResponse,
)


class ResponseInterrupt(Exception):
    """ To use when the response is asynchronous
        or infinite until top level wants to finish it.
    """


def ac_rx_data(self, event):
    """ Receiving data from the partner clisrv gsock.
        Can be one or more requests.
    """
    self.start_inactivity_timer()
    #gsock = event.source[-1]
    gsock = self.gsock  # must be same as event.source[-1]
    data = event.data
    if not data:
        return
    new_request = self.parsing_request
    while data:
        if new_request is None:
            new_request = HTTPRequestParser(self)
        n = new_request.received(data)
        if new_request.expect_continue and new_request.headers_finished:
            # guaranteed by parser to be a 1.1 new_request
            new_request.expect_continue = False

            if not self.sent_continue:
                self.send_event(
                    gsock,
                    'EV_SEND_DATA',
                    data=b'HTTP/1.1 100 Continue\r\n\r\n'
                )
                self.sent_continue = True
                new_request.completed = False

        if new_request.completed:
            # The new_request (with the body) is ready to use.
            self.parsing_request = None
            if not new_request.empty:
                self.enqueue_request(new_request)
            new_request = None
        else:
            self.parsing_request = new_request
        if n >= len(data):
            break
        data = data[n:]

    if len(self.dl_requests):
        self.send_event(self, 'EV_DEQUEUE_REQUEST')


def ac_dequeue_request(self, event):
    if self.responding_request:
        self.logger.exception('Internal ERROR!!!: '
                              'responding_request MUST be None')
    if len(self.dl_requests):
        self.responding_request = self.dl_requests.popleft()
        self.send_event(self, 'EV_HTTP_REQUEST')
    else:
        self.start_inactivity_timer()


def ac_http_request(self, event):
    """ Internal event.
        The request is saved in self.responding_request.
    """
    self.stop_inactivity_timer()

    if self.responding_request.error:
        response = HttpErrorResponse(self.responding_request)
        response.execute()
        self.finish(response)
        return

    self.start_responseless_timer()
    self.set_new_state('ST_WAIT_RESPONSE')
    # TODO: in stratus environment, we need to inform of who srvcli is.
    self.broadcast_event(
        'EV_HTTP_REQUEST',
        request=self.responding_request,
        channel=self,
    )


def ac_http_response(self, event):
    response = event.response
    if not isinstance(response, HttpResponse):
        self.logger and self.logger.error(
            "ERROR response doesn't mach HttpResponse %s" %
            response.request.path)

    if response.request != self.responding_request:
        self.logger and self.logger.error(
            "ERROR response doesn't mach responding request %s" %
            response.request.path)

    if self.responding_response:
        self.logger and self.logger.error(
            "ERROR responding_response is BUSY, of %s" %
            response.request.path)
    self.responding_response = response

    response.start()
    try:
        response.execute()
    except ResponseInterrupt:
        """ Response is asynchronous or infinite.
            Don't clear the current responding_request.
        """
        self.stop_responseless_timer()  # TODO: do some ping-alive
        return

    except:
        self.logger and self.logger.exception(
            'Exception when serving %s' % response.request.path)
        if not response.wrote_header:
            if self.parent.config.expose_tracebacks:
                body = traceback.format_exc()
            else:
                body = ('The server encountered an unexpected '
                        'internal server error')

            request = HTTPRequestParser(self)
            request.error = InternalServerError(body)
            response = HttpErrorResponse(request)
            response.execute()
        else:
            response.close_on_finish = True

    self.finish(response)


def ac_transmit_ready(self, event):
    pass


def ac_inactivity_timeout(self, event):
    """ Close the channel by inactivity.
    """
    if self.gsock:
        self.send_event(self.gsock, 'EV_DROP')


def ac_responseless_timeout(self, event):
    """ Close the channel by responseless of top level.
    """
    if not self.gsock:
        return

    body = (
        'Response Timeout. The server is busy. '
        'Please re-try your request in a few moments.'
    )
    request = HTTPRequestParser(self)
    request.error = InternalServerError(body)
    response = HttpErrorResponse(request)
    response.execute()
    self.finish(response)


def ac_disconnected(self, event):
    """ Gsock closed.
    """
    self.broadcast_event('EV_HTTP_CHANNEL_CLOSED', channel=self)


GHTTPCLISRV_FSM = {
    'event_list': (
        'EV_SET_TIMER: bottom output',
        'EV_RESPONSELESS_TIMEOUT: bottom input',
        'EV_INACTIVITY_TIMEOUT: bottom input',
        'EV_DISCONNECTED: bottom input',
        'EV_RX_DATA: bottom input',
        'EV_DEQUEUE_REQUEST',
        'EV_TRANSMIT_READY: bottom input',
        'EV_SEND_DATA: bottom output',
        'EV_FLUSH_OUTPUT_DATA: bottom output',
        'EV_WRITE_OUTPUT_DATA: bottom output',
        'EV_HTTP_CHANNEL_OPENED: top output',
        'EV_HTTP_CHANNEL_CLOSED: top output',
        'EV_HTTP_REQUEST: top output',
        'EV_HTTP_RESPONSE: top input',
    ),
    'state_list': (
        'ST_IDLE',
        'ST_WAIT_RESPONSE',
    ),
    'machine': {
        'ST_IDLE':
        (
            ('EV_DISCONNECTED',         ac_disconnected,            None),
            ('EV_INACTIVITY_TIMEOUT',   ac_inactivity_timeout,      None),
            ('EV_RX_DATA',              ac_rx_data,                 None),
            ('EV_DEQUEUE_REQUEST',      ac_dequeue_request,         None),
            ('EV_HTTP_REQUEST',         ac_http_request,            None),
        ),
        'ST_WAIT_RESPONSE':
        (
            ('EV_DISCONNECTED',         ac_disconnected,            None),
            ('EV_RESPONSELESS_TIMEOUT', ac_responseless_timeout,    'ST_IDLE'),
            ('EV_RX_DATA',              ac_rx_data,                 None),
            ('EV_DEQUEUE_REQUEST',      None,                       None),
            ('EV_HTTP_RESPONSE',        ac_http_response,           'ST_IDLE'),

            ('EV_TRANSMIT_READY',       ac_transmit_ready,          None),
        ),
    }
}

GHTTPCLISRV_GCONFIG = {
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
    ],
    'gsock': [
        None, None, GConfig.FLAG_DIRECT_ATTR, None, "partner gsock."
    ],
    'maximum_simultaneous_requests': [
        int, 0, 0, None,
        "maximum simultaneous requests."
    ],
    'inactivity_timeout': [
        int, 5 * 60 * 60, 0, None,
        "Inactivity timeout in seconds."
    ],
    'responseless_timeout': [
        int, 5 * 60 * 60, 0, None,
        "'Without response' timeout in seconds."
    ],

    # Used in parser:
    'url_scheme': [str, 'http', 0, None, "default ``http`` value"],
    # A tempfile should be created if the pending input is larger than
    # inbuf_overflow, which is measured in bytes. The default is 512K.  This
    # is conservative.
    'inbuf_overflow': [int, 524288, 0, None, ""],
    # maximum number of bytes of all request headers combined (256K default)
    'max_request_header_size': [int, 262144, 0, None, ""],
    # maximum number of bytes in request body (1GB default)
    'max_request_body_size': [int, 1073741824, 0, None, ""],
}


class GHttpCliSrv(GObj):
    """  Http clisrv (client of server) class.

    This gobj is create by GHttpServer when it receives an EV_CONNECTED event
    from a new gsock gobj.

    This class will subscribe all the events of the partner
    :class:`ginsfsm.c_sock.GSock` gobj, to implement the http protocol.

    .. ginsfsm::
       :fsm: GHTTPCLISRV_FSM
       :gconfig: GHTTPCLISRV_GCONFIG

    *Top Output-Events:*
        * :attr:`'EV_HTTP_CHANNEL_OPENED'`: new http client.

          Event attributes:

            * ``channel``: http channel.

        * :attr:`'EV_HTTP_CHANNEL_CLOSED'`: http client closed.

          Event attributes:

            * ``channel``: http channel.

        * :attr:`'EV_HTTP_REQUEST'`: new http request.

          Event attributes:

            * ``channel``: http channel.
            * ``request``: http request.

    *Top Input-Events:*
        * :attr:`'EV_HTTP_RESPONSE'`: response to the current request.

          Event attributes:

            * ``response``: http response.

    *Bottom Input-Events:*
        * :attr:`'EV_DISCONNECTED'`: socket disconnected.

          The clisrv `gobj` will be destroyed.

        * :attr:`'EV_TRANSMIT_READY'`: socket ready to transmit more data.

        * :attr:`'EV_RX_DATA'`: data received. Process http protocol.

    *Bottom Output-Events:*

        * :attr:`'EV_SEND_DATA'`: transmit data socket.
        * :attr:`'EV_WRITE_OUTPUT_DATA'`: write data to socket output buffer.
        * :attr:`'EV_FLUSH_OUTPUT_DATA'`: flush data of socket output buffer.

    """

    def __init__(self):
        GObj.__init__(self, GHTTPCLISRV_FSM, GHTTPCLISRV_GCONFIG)
        self.parsing_request = None  # A request parser instance
        self.responding_request = None  # request waiting a top response
        self.responding_response = None  # current response being responding
        self.dl_requests = deque()  # requests queue
        self.sent_continue = False  # used as a latch after sending 100continue

    def go_out(self):
        """ Finish zone.
        """

    def start_up(self):
        if self.config.subscriber is None:
            self.config.subscriber = self.parent

        # gsock MUST be passed at creating gobj

        # Canalize the flow of messages
        # bottom events for me
        self.gsock.subscribe_event(None, self)

        # we need to know disconnected event for deleting gobjs
        self.gsock.subscribe_event(
            'EV_DISCONNECTED',
            self,
            __hard_subscription__=True,
        )

        # top events for subscriber
        self.subscribe_event(None, self.config.subscriber)
        self.broadcast_event('EV_HTTP_CHANNEL_OPENED', channel=self)

        # Setup the timers
        self.inactivity_timer = self.create_gobj(
            None,
            GTimer,
            self,
            timeout_event_name='EV_INACTIVITY_TIMEOUT'
        )
        self.start_inactivity_timer()

        self.responseless_timer = self.create_gobj(
            None,
            GTimer,
            self,
            timeout_event_name='EV_RESPONSELESS_TIMEOUT'
        )

    def enqueue_request(self, new_request):
        self.dl_requests.append(new_request)
        max_req = self.config.maximum_simultaneous_requests
        if max_req > 0 and len(self.dl_requests) > max_req:
            # Close the channel by maximum simultaneous requests reached.
            body = 'Please change your behavior.' \
                ' You have reached the maximum simultaneous requests (%d).' % (
                    max_req)
            request = HTTPRequestParser(self)
            request.error = InternalServerError(body)
            response = HttpErrorResponse(request)
            response.service()
            self.clear_request_queue()

    def clear_request_queue(self):
        for request in self.dl_requests:
            request._close()
        self.dl_requests.clear()

    def start_inactivity_timer(self):
        self.send_event(
            self.inactivity_timer,
            'EV_SET_TIMER',
            seconds=self.config.inactivity_timeout
        )

    def stop_inactivity_timer(self):
        self.send_event(
            self.inactivity_timer,
            'EV_SET_TIMER',
            seconds=-1
        )

    def start_responseless_timer(self):
        self.send_event(
            self.responseless_timer,
            'EV_SET_TIMER',
            seconds=self.config.responseless_timeout
        )

    def stop_responseless_timer(self):
        self.send_event(
            self.responseless_timer,
            'EV_SET_TIMER',
            seconds=-1
        )

    def write(self, data):
        """ Write data to output buffer.
            To supply asynchronous access to high level.
        """
        if not self.responding_response:
            self.logger and self.logger.error(
                "ERROR channel.write() with no responding_response")
            return
        self.responding_response.write(data)

    def flush(self, callback=None):
        """ Flush output buffer.
            To supply asynchronous access to high level.
        """
        if not self.responding_response:
            self.logger and self.logger.error(
                "ERROR channel.flush() with no responding_response")
            return
        self.responding_response.flush(callback)

    def finish(self, response=None):
        """ Finishes this response,
            flushing output buffer,
            and ending the HTTP request.
            To supply asynchronous access to high level.
        """
        if response is None:
            if not self.responding_response:
                self.logger and self.logger.error(
                    "ERROR channel.finish()"
                    " with no responding_response")
                return
            response = self.responding_response
        response.finish()
        self.stop_responseless_timer()
        self.responding_request = None
        self.responding_response = None

        if response.close_on_finish:
            # ignore all enqueued requests.
            self.clear_request_queue()
            if self.gsock:
                self.gsock.mt_drop()
            return

        # pull the request queue
        self.post_event(self, 'EV_DEQUEUE_REQUEST')



# -*- encoding: utf-8 -*-
"""
GObj :class:`WsgiResponse`
===========================

Code copied from waitress.task.py adapted to ginsfsm.

.. autoclass:: WsgiResponse
    :members:

"""

from ginsfsm.buffers import ReadOnlyFileBasedBuffer
from ginsfsm.compat import (
    reraise,
)
from ginsfsm.protocols.http.common.parser import build_environment
from ginsfsm.protocols.http.common.response import HttpResponse

hop_by_hop = frozenset((
    'connection',
    'keep-alive',
    'proxy-authenticate',
    'proxy-authorization',
    'te',
    'trailers',
    'transfer-encoding',
    'upgrade',
))


class WsgiResponse(HttpResponse):
    """Get a response from a WSGI application.
    """
    environ = None

    def __init__(self, request, wsgi_server, application):
        HttpResponse.__init__(self, request)
        self.wsgi_server = wsgi_server
        self.application = application

    def execute(self):
        env = self.get_wsgi_environment()

        def start_response(status, headers, exc_info=None):
            if self.complete and not exc_info:
                raise AssertionError("start_response called a second time "
                                     "without providing exc_info.")
            if exc_info:
                try:
                    if self.complete:
                        # higher levels will catch and handle raised exception:
                        # 1. "service" method in task.py
                        # 2. "service" method in gsock.py
                        # 3. "handler_thread" method in task.py
                        reraise(exc_info[0], exc_info[1], exc_info[2])
                    else:
                        # As per WSGI spec existing headers must be cleared
                        self.response_headers = []
                finally:
                    exc_info = None

            self.complete = True

            if not status.__class__ is str:
                raise AssertionError('status %s is not a string' % status)

            self.status = status

            # Prepare the headers for output
            for k, v in headers:
                if not k.__class__ is str:
                    raise AssertionError(
                        'Header name %r is not a string in %r' % (k, (k, v))
                        )
                if not v.__class__ is str:
                    raise AssertionError(
                        'Header value %r is not a string in %r' % (v, (k, v))
                        )
                kl = k.lower()
                if kl == 'content-length':
                    self.content_length = int(v)
                elif kl in hop_by_hop:
                    raise AssertionError(
                        '%s is a "hop-by-hop" header; it cannot be used by '
                        'a WSGI application (see PEP 3333)' % k)

            self.response_headers.extend(headers)

            # Return a method used to write the response data.
            return self.write

        # Call the application to handle the request and return a response
        app_iter = self.application(env, start_response)

        try:
            if app_iter.__class__ is ReadOnlyFileBasedBuffer:
                cl = self.content_length
                size = app_iter.prepare(cl)
                if size:
                    if cl != size:
                        if cl is not None:
                            self.remove_content_length_header()
                        self.content_length = size
                    self.write(b'')  # generate headers
                    self.gsock.mt_write(app_iter)
                    return

            first_chunk_len = None
            for chunk in app_iter:
                if first_chunk_len is None:
                    first_chunk_len = len(chunk)
                    # Set a Content-Length header if one is not supplied.
                    # start_response may not have been called until first
                    # iteration as per PEP, so we must reinterrogate
                    # self.content_length here
                    if self.content_length is None:
                        app_iter_len = None
                        if hasattr(app_iter, '__len__'):
                            app_iter_len = len(app_iter)
                        if app_iter_len == 1:
                            self.content_length = first_chunk_len
                # transmit headers only after first iteration of the iterable
                # that returns a non-empty bytestring (PEP 3333)
                if chunk:
                    self.write(chunk)

            cl = self.content_length
            if cl is not None:
                if self.content_bytes_written != cl:
                    # close the connection so the client isn't sitting around
                    # waiting for more data when there are too few bytes
                    # to service content-length
                    self.close_on_finish = True
                    if self.request.command != 'HEAD':
                        self.wsgi_server.logger.error(
                            'application returned too few bytes (%s) '
                            'for specified Content-Length (%s) via app_iter' % (
                                self.content_bytes_written, cl),
                            )
        finally:
            if hasattr(app_iter, 'close'):
                app_iter.close()

    def get_wsgi_environment(self):
        """Returns a WSGI environment."""
        environ = self.environ
        if environ is not None:
            # Return the cached copy.
            return environ

        request = self.request
        channel = request.channel
        gsock = request.channel.gsock
        wsgi_server = self.wsgi_server

        server_name = wsgi_server.server_name
        server_port = str(wsgi_server.effective_port)
        remote_addr = gsock.addr[0]
        environ = build_environment(
            request,
            server_name,
            server_port,
            remote_addr
        )
        # ginsfsm variables to do asynchronous response
        # you can access directly to http channel with the next functions:
        #   write(data):        Write data to output buffer.
        #   flush(callback):    Flush output buffer to socket,
        #                       and call callback when transmit ready event.
        #   finish()            Finish http response.
        environ['ginsfsm.channel'] = channel
        environ['ginsfsm.gsock'] = gsock

        self.environ = environ
        return environ



"""
gconsole
========

Utility for connect to inside of GAplics

"""

from code import InteractiveInterpreter
import sys
import argparse
import logging
logging.basicConfig(level=logging.INFO)


from ginsfsm.utils import random_key
from ginsfsm.gobj import GObj
from ginsfsm.gaplic import GAplic
from ginsfsm.smachine import SMachine
from ginsfsm.c_timer import GTimer
from ginsfsm.c_sock import GSock
from ginsfsm.utils import hexdump
from ginsfsm.c_stdin import GStdin

my_commands = [
    'help',
    'gaplic',
    'router',
    'set_global_trace',
    'show_routes',
    'add_route',
    'delete_route',
    'show_external_routes',
    'external_command',
    # remember add command help to env_help
]


def ac_rx_data(self, event):
    data = event.kw['data']
    if 0 or self.config.verbose:
        print('Receiving %d bytes' % len(data))
        print(hexdump('<=', data))
    if not data:
        self.stdout.write('\n>>> ')
        self.stdout.flush()
        return

    first_word = data.split()[0]
    if first_word in my_commands:
        command = 'cmd_' + first_word
        if hasattr(self, command):
            cmd = getattr(self, command)
            params = data.split()[1:]
            s = cmd(*params)
            if s:
                self.stdout.write(s)
                self.stdout.flush()
        else:
            self.stdout.write('ERROR: command %r NOT implemented.\n' % command)
    else:
        self.interpreter.runsource(data)

    self.stdout.write('\n>>> ')
    self.stdout.flush()


def ac_timeout(self, event):
    """
    """
    #self.set_timeout(5)


def ac_command_response(self, event):
    """
    """
    #self.set_timeout(5)
    action_return = event.action_return
    self.stdout.write(action_return)
    self.stdout.flush()
    self.stdout.write('\n>>> ')
    self.stdout.flush()


GCONSOLE_FSM = {
    'event_list': (
        'EV_SET_TIMER: bottom output',
        'EV_TIMEOUT:bottom input',
        'EV_RX_DATA:bottom input',
        'EV_COMMAND',
        'EV_SHOW_ROUTES',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_TIMEOUT',      ac_timeout,             'ST_IDLE'),
            ('EV_RX_DATA',      ac_rx_data,             'ST_IDLE'),
            ('EV_COMMAND',      ac_command_response,    'ST_IDLE'),
            ('EV_SHOW_ROUTES',  ac_command_response,    'ST_IDLE'),
        ),
    }
}


GCONSOLE_GCONFIG = {  # type, default_value, flag, validate_function, desc
    'verbose': [int, 0, 0, None, "Increase output verbosity. Values [0,1,2]"],
}


class GConsole(GObj):
    """  GConsole GObj.

    .. ginsfsm::
       :fsm: GCONSOLE_FSM
       :gconfig: GCONSOLE_GCONFIG

    *Input-Events:*

        * :attr:`'EV_TIMEOUT'`: Timer over.
            Start the machine.

        * :attr:`'EV_RX_DATA'`: Receiving data.
            Receiving data from urls.

    *Output-Events:*

        * :attr:`'EV_START_TIMER'`: Start timer.

    """
    def __init__(self):
        GObj.__init__(self, GCONSOLE_FSM, GCONSOLE_GCONFIG)

    def start_up(self):
        """ Initialization zone."""
        self.stdout = sys.stdout
        self.timer = self.create_gobj(
            None,
            GTimer,
            self,
        )

        if self.config.verbose > 1:
            #TODO: move to local_conf
            settings = {
                'GObj.trace_mach': True,
                'GObj.logger': logging,
            }
            self.overwrite_parameters(-1, **settings)

        self.create_gobj(
            None,
            GStdin,
            self,
        )

        # setup help text for default environment
        env_help = dict()
        env_help['1. help'] = 'Show this help.'
        env_help['2. gaplic'] = 'This gaplic instance.'
        env_help['3. SMachine'] = 'SMachine class.'
        env_help['4. set_global_trace'] = '<bool> Set global trace.'
        env_help['5. show_routes'] = 'Show routes.'
        env_help['6. add_route'] = '<url> (Connect to external gaplic).'
        env_help['7. delete_route'] = '<route_ref> (Remove route).'
        env_help['8. show_external_routes'] = \
            '<route_ref> (Show routes of external gaplic linked by route_ref).'
        env_help['9. external_command'] = \
            '<route_ref> <command> (Execute command in external gaplic linked by route_ref).'

        env = {
            '__name__': "__ginsfsm_console__",
            '__doc__': 'GinsFSM console',
            'gaplic': self.gaplic,
            'SMachine': SMachine,
        }
        self.interpreter = InteractiveInterpreter(locals=env)

        # generate help text
        help = ''
        if env_help:
            help += 'Available commands:'
            for var in sorted(env_help.keys()):
                help += '\n  %-12s %s' % (var, env_help[var])
            help += '\n'
        self.shelp = help

        cprt = 'GinsFSM console.'
        banner = "Python %s on %s\n%s" % (sys.version, sys.platform, cprt)
        banner += '\n\n' + help + '>>> '
        self.stdout.write(banner)
        self.stdout.flush()

        self.set_timeout(1)

    def set_timeout(self, seconds):
        self.send_event(self.timer, 'EV_SET_TIMER', seconds=seconds)

    def clear_timeout(self):
        self.send_event(self.timer, 'EV_SET_TIMER', seconds=-1)

    def cmd_help(self):
        return self.shelp

    def cmd_gaplic(self):
        return repr(self.gaplic)

    def cmd_set_global_trace(self, value):
        """  parameter always become as strings
        """
        value = eval(value)
        SMachine.global_trace_mach(value)
        GSock.global_trace_dump(value)
        logger = logging.getLogger()
        if value:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

    def cmd_show_routes(self):
        router = self.gaplic.router
        ret = self.send_event(
            router,
            'EV_SHOW_ROUTES',
        )
        return ret

    def cmd_add_route(self, url):
        """
        http://localhost:8000/apihost10/__pyramid_router__/websocket
        """
        router = self.gaplic.router
        ret = self.send_event(
            router,
            'EV_ADD_STATIC_ROUTE',
            name=None,
            roles=None,
            urls=url,
        )
        return repr(ret)

    def cmd_delete_route(self, route_ref):
        registry = self.gaplic.router.registry
        route = registry.get_route_by_ref(route_ref)
        if not route:
            return 'Route NOT found'
        if registry.delete_route(route_ref):
            return 'Route deleted'
        return 'Route CANNOT be deleted'

    def cmd_show_external_routes(self, route_ref):
        self.gaplic.router.mt_send_event_to_external_route(
            route_ref,
            'router',
            'EV_SHOW_ROUTES',  # 'EV_COMMAND',
            self.name,
            {
                '__subscribe_response__': True,
            },
        )

    def cmd_external_command(self, route_ref, command):
        self.gaplic.router.mt_send_event_to_external_route(
            route_ref,
            'router',
            'EV_COMMAND',
            self.name,
            {
                '__subscribe_response__': True,
                'command': command,
            },
        )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase output verbosity",
        type=int,
        choices=[0, 1, 2],
        default=1,
    )

    local_conf = {
        'router_enabled': True,
        'GRouter.server': True,
        'GRouter.trace_router': True,
        'GWebSocket.trace_mach': False,
        'GSock.trace_dump': False,
        'GObj.trace_mach': False,
        'GObj.logger': logging,
    }
    ga = GAplic(name=random_key(), roles='gcmd',  **local_conf)
    ga.create_gobj(
        'gconsole',
        GConsole,
        ga,
        __unique_name__=True,
    )

    try:
        ga.start()
    except (KeyboardInterrupt, SystemExit):
        print('Program stopped')


if __name__ == "__main__":
    main()



"""
`Sockjs <https://github.com/sockjs>`_ server for
`ginsfsm <http://ginsfsm.org>`_
and
`Pyramid <http://docs.pylonsproject.org/en/latest/index.html>`_.

"""


def includeme(config):
    """Include the ginsfsm sockjs configuration
    """
    config.scan("ginsfsm.protocols.sockjs.server")



from PIL import ImageFont

colors = [
    '#edd400',   # butter 2
    '#73d216',   # chameleon 2
    '#fcaf3e',   # orange 1
    '#729fcf',   # sky blue 1
    '#ad7fa8',   # plum 1
    '#e9b96e',   # chocolate 1
    '#ef2929',   # scarlet red 1
    #'#d3d7cf',   # aluminium 2
    '#c4a000',   # butter 3
    '#4e9a06',   # chameleon 3
    '#f57900',   # orange 2
    '#3465a4',   # sky blue 2
    '#75507b',   # plum 2
    '#c17d11',   # chocolate 2
    '#a40000',   # scarlet red 3
    '#888a85',   # aluminium 4
]


def right_arrow_path(x, y, offset):
    """ x,y is the final point
    """
    d = [('m', x - 2 * offset, y),
        (offset, 0),
        (0, -offset),
        (offset, offset),
        (-offset, offset),
        (0, -offset),
        ('z')
    ]
    return d


def left_arrow_path(x, y, offset):
    """ x,y is the final point
    """
    d = [('m', x + 2 * offset, y),
        (-offset, 0),
        (0, offset),
        (-offset, -offset),
        (offset, -offset),
        (0, offset),
        ('z')
    ]
    return d


def top_arrow_path(x, y, offset):
    """ x,y is the final point
    """
    d = [('m', x, y + 2 * offset),
        (0, -offset),
        (-offset, 0),
        (offset, -offset),
        (offset, offset),
        (-offset, 0),
        ('z')
    ]
    return d


def bottom_arrow_path(x, y, offset):
    """ x,y is the final point
    """
    d = [('m', x, y - 2 * offset),
        (0, offset),
        (offset, 0),
        (-offset, offset),
        (-offset, -offset),
        (offset, 0),
        ('z')
    ]
    return d


class Point(tuple):
    mapper = dict(x=0, y=1)

    def __new__(cls, x, y):
        return super(Point, cls).__new__(cls, (x, y))

    def __getattr__(self, name):
        return self[self.mapper[name]]

    def shift(self, x=0, y=0):
        return self.__class__(self.x + x, self.y + y)


class Line(tuple):
    mapper = dict(x1=0, y1=1, x2=2, y2=3)

    def __new__(cls, x1, y1, x2, y2):
        return super(Line, cls).__new__(cls, (x1, y1, x2, y2))

    def __getattr__(self, name):
        return self[self.mapper[name]]

    def shift(self, x=0, y=0):
        return self.__class__(
            self.x1 + x, self.y1 + y, self.x2 + x, self.y2 + y)

    @property
    def start(self):
        return Point(self.x1, self.y1)

    @property
    def end(self):
        return Point(self.x2, self.y2)


class Size(tuple):
    mapper = dict(cx=0, cy=1)

    def __new__(cls, cx, cy):
        return super(Size, cls).__new__(cls, (cx, cy))

    def __getattr__(self, name):
        return self[self.mapper[name]]

    def shift(self, cx=0, cy=0):
        return self.__class__(self.cx + cx, self.cy + cy)

    def extend(self, cx=0, cy=0):
        return self.__class__(self.cx + cx, self.cy + cy)

    @property
    def width(self):
        return self.cx

    @property
    def height(self):
        return self.cy


class Box(list):
    mapper = dict(x=0, y=1, cx=2, cy=3)

    def __init__(self, x, y, cx, cy):
        super(Box, self).__init__((x, y, cx, cy))

    def __getattr__(self, name):
        return self[self.mapper[name]]

    def __repr__(self):
        class_name = self.__class__.__name__
        x = self.x
        y = self.y
        width = self.width
        height = self.height
        addr = id(self)

        format = "<%(class_name)s (%(x)s, %(y)s) " + \
                 "%(width)dx%(height)d at 0x%(addr)08x>"
        return format % locals()

    def shift(self, x=0, y=0):
        return self.__class__(self.x + x, self.y + y,
                              self.cx, self.cy)

    def extend(self, cx=0, cy=0):
        return self.__class__(self.x, self.y,
                              self.cx + cx, self.cy + cy)

    @property
    def xy(self):
        return Point(self.x, self.y)

    @property
    def size(self):
        return Size(self.cx, self.cy)

    @property
    def width(self):
        return self.cx

    @property
    def height(self):
        return self.cy

    @property
    def topleft(self):
        return Point(self.x, self.y)

    @property
    def topright(self):
        return Point(self.x + self.cx, self.y)

    @property
    def bottomleft(self):
        return Point(self.x, self.y + self.cy)

    @property
    def bottomright(self):
        return Point(self.x + self.cx, self.y + self.cy)


def get_textsize(text, font_name, font_size):
    """ Return a tuple with
    """
    font = ImageFont.truetype(font_name, font_size)
    size = font.getsize(text)
    ascent, descent = font.getmetrics()
    return Size(size[0], size[1]), ascent, descent




# -*- encoding: utf-8 -*-
"""
    SockJS session implementation.

    All transport urls come with /{server_id}/{session_id} subpath.

    The :class:GSockjsSession and :class:GSessionId classes
    captures these segments.

"""
from ginsfsm.gobj import GObj
from ginsfsm.gconfig import GConfig
from ginsfsm.protocols.sockjs.server.c_transport_xhr import (
    GXhrSend,
    GXhrPolling,
    GXhrStreaming,
)
from ginsfsm.protocols.sockjs.server.c_transport_jsonp import (
    GJsonpSend,
    GJsonpPolling,
)
from ginsfsm.protocols.sockjs.server.c_transport_eventsource import (
    GEventsourceStreaming,
)
from ginsfsm.protocols.sockjs.server.c_transport_htmlfile import (
    GHtmlfileStreaming,
)
from ginsfsm.protocols.sockjs.server.c_transport_websocket import (
    GWebsocket,
)


#----------------------------------------------------------------#
#                   GSockjsSession GClass
#----------------------------------------------------------------#
GSOCKJSSESSION_GCONFIG = {
    'sockjs_server': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ""],
}


class GSockjsSession(GObj):
    """  GSockjsSession GObj aka GServerID.
    This gobj treats the {server_id} segment of the url.
    The main reason for this segment is to make it easier
    to configure load balancer
    and enable sticky sessions based on first part of the url.
    The sockjs serve must ignore it.

    This gobj would be called GServerID instead of GSockjsSession.
    We don't name this gobj as GServerID for a better understanding of the
    parent gobj view point.
    """
    def __init__(self):
        GObj.__init__(self, {}, GSOCKJSSESSION_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        self.session_id = self.create_gobj(
            'session_id',
            GSessionId,
            self,
            re_name='^[^/.]+$',
            sockjs_server=self.sockjs_server,
        )

#----------------------------------------------------------------#
#                   GSessionId GClass
#----------------------------------------------------------------#
GSESSIONID_GCONFIG = {
    'sockjs_server': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ""],
}


class GSessionId(GObj):
    """  GSessionId GObj.
    """
    def __init__(self):
        GObj.__init__(self, {}, GSESSIONID_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        #-------------------------------------------#
        #           Global handlers
        #               /xhr_send
        #               /jsonp_send
        #-------------------------------------------#
        self.create_gobj(
            'xhr_send',
            GXhrSend,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'jsonp_send',
            GJsonpSend,
            self,
            sockjs_server=self.sockjs_server,
        )

        #-------------------------------------------#
        #           Transports
        #               /websocket
        #               /xhr
        #               /xhr_streaming
        #               /jsonp
        #               /eventsource
        #               /htmlfile
        #-------------------------------------------#
        self.create_gobj(
            'websocket',
            GWebsocket,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'xhr',
            GXhrPolling,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'xhr_streaming',
            GXhrStreaming,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'jsonp',
            GJsonpPolling,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'eventsource',
            GEventsourceStreaming,
            self,
            sockjs_server=self.sockjs_server,
        )
        self.create_gobj(
            'htmlfile',
            GHtmlfileStreaming,
            self,
            sockjs_server=self.sockjs_server,
        )



# -*- encoding: utf-8 -*-

import os
import sys
import banco

from setuptools import setup, find_packages

if sys.version_info[:2] < (2, 6):
    raise RuntimeError('Requires Python 2.6 or better')

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.txt')).read()
except IOError:
    README = CHANGES = ''

requires = [''
    'ginsfsm',
    'pyramid >= 1.4',
]

setup(
    name='banco',
    version=banco.__version__,
    description='banco',
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        "Programming Language :: Python",
    ],
    author='',
    author_email='',
    url='',
    license='MIT License',
    keywords='gobj ginsfsm finite state machine fsm wsgi',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    tests_require=requires,
    test_suite="banco.tests",
    entry_points="""\
    [paste.app_factory]
    main = banco.main:main
    pyramid_app = banco.main:pyramid_paste_factory
    """,
)



# -*- coding: utf-8 -*-
"""
    SockJS connection interface with GOBJ
"""
from ginsfsm.gobj import GObj


def ac_on_open(self, event):
    pass


def ac_on_close(self, event):
    pass


def ac_on_message(self, event):
    """ Message received.
    """
    msg = event.msg
    self.send(msg)


SOCKJSCONNECTION_FSM = {
    'event_list': (
        'EV_ON_OPEN:bottom input',
        'EV_ON_MESSAGE:bottom input',
        'EV_ON_CLOSE:bottom input',
    ),
    'state_list': (
        'ST_IDLE',
    ),
    'machine': {
        'ST_IDLE':
        (
            ('EV_ON_OPEN',      ac_on_open,     None),
            ('EV_ON_MESSAGE',   ac_on_message,  None),
            ('EV_ON_CLOSE',     ac_on_close,    None),
        ),
    }
}

SOCKJSCONNECTION_GCONFIG = {
    'session': [None, None, 0, None, ""],
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
        "Default is ``None``, i.e., the parent"
    ],
}


class SockJSConnection(GObj):
    """  SockJSConnection GObj.

    .. ginsfsm::
       :fsm: SOCKJSCONNECTION_FSM
       :gconfig: SOCKJSCONNECTION_GCONFIG

    *Input-Events:*
        * :attr:`'EV_ON_OPEN'`: sockjs client connected.

        * :attr:`'EV_ON_MESSAGE'`: message received.

          Event attributes:
            * :attr:`msg`: the message received.

        * :attr:`'EV_ON_CLOSE'`: sockjs client disconnected.



    """
    def __init__(self):
        GObj.__init__(self, SOCKJSCONNECTION_FSM, SOCKJSCONNECTION_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        if self.config.subscriber is None:
            self.config.subscriber = self.parent
        self.subscribe_event(None, self.config.subscriber)

    def send(self):
        """ Transmit message to sockjs client.
        """
        self.session.send()



# -*- encoding: utf-8 -*-
"""
Pyramid root entry point

.. autoclass:: PyramidRoot
    :members:

:func:`pyramid_application`.

"""

from pyramid.config import Configurator
from pyramid.security import (
    Allow,
    Everyone,
    ALL_PERMISSIONS,
)

from ginsfsm.globals import get_global_app
from ginsfsm.gobj import GObj
from .pyramid_main import TestSocksjsApps


#----------------------------------------------------------------------#
#                           Pyramid root
#----------------------------------------------------------------------#
PYRAMIDROOT_GCONFIG = {
}


class PyramidRoot(GObj):
    """ Pyramid root.
        Must be a gobj with an empty name.
    """
    __acl__ = [
        (Allow, Everyone, ALL_PERMISSIONS)
    ]

    def __init__(self):
        GObj.__init__(self, {}, PYRAMIDROOT_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """
        self.pyramid_app = self.create_gobj(
            '__test_sockjs__',
            TestSocksjsApps,
            self,
        )

pyramid_root = None


def make_root(gaplic):
    global pyramid_root
    if pyramid_root is None:
        pyramid_root = gaplic.create_gobj(
            None,   # pyramid resources tree root MUST be unnamed.
            PyramidRoot,
            None,
        )
    return pyramid_root


def get_root(request):
    return pyramid_root


#-------------------------------------------------#
#       Paste entry point
#       Pyramid wsgi-app configuration
#-------------------------------------------------#
def pyramid_application(global_config, **local_conf):
    """ Paste entry point.

        Returns a WSGI application.
    """
    if 'gaplic' in local_conf:
        gaplic_name = local_conf.pop('gaplic')
    else:
        raise Exception('You must supply a gaplic name in ini file.')

    gaplic = get_global_app(gaplic_name)
    if not gaplic:
        raise Exception("gaplic '%s' supplied for wsgi is invalid" %
                        gaplic_name)

    config = Configurator(
        root_factory=get_root,
        settings=local_conf,
    )

    # Set up views
    config.include("ginsfsm.protocols.sockjs.server")

    make_root(gaplic)

    return config.make_wsgi_app()



# -*- coding: utf-8 -*-
"""
    EventSource transport implementation.
"""
import logging
from pyramid.view import view_config

from ginsfsm.gobj import GObj
from ginsfsm.gconfig import GConfig
from ginsfsm.protocols.sockjs.server.basehandler import SessionHandler

#----------------------------------------------------------------#
#                   GEventsourceStreaming GClass
#                   /eventsource
#----------------------------------------------------------------#
GEVENTSOURCESTREAMING_GCONFIG = {
    'sockjs_server': [None, None, GConfig.FLAG_DIRECT_ATTR, None, ""],
}


class GEventsourceStreaming(GObj):
    """  GEventsourceStreaming GObj.
    """
    def __init__(self):
        GObj.__init__(self, {}, GEVENTSOURCESTREAMING_GCONFIG)

    def start_up(self):
        """ Initialization zone.
        """


#----------------------------------------------------------------#
#                   GEventsourceStreaming Views
#                   /eventsource
#----------------------------------------------------------------#
@view_config(
    context=GEventsourceStreaming,
    name='',
    attr='options',
    request_method='OPTIONS',
)
@view_config(
    context=GEventsourceStreaming,
    name='',
    attr='get',
    request_method='GET',
)
class EventsourceStreamingTransport(SessionHandler):
    name = 'eventsource'

    def __init__(self, context, request):
        super(EventsourceStreamingTransport, self).__init__(
            context,
            request,
            True  # Asynchronous response!
        )
        self.session = None
        self.active = True
        self.amount_limit = self.context.sockjs_server.config.response_limit

    def get(self):
        response = self.response

        session_id = self.sid = self.context.parent.re_matched_name

        # Start response
        self.preflight()
        self.handle_session_cookie()
        self.disable_cache()

        response.content_type = 'text/event-stream'
        response.charset = 'UTF-8'
        self.write('\r\n')
        self.flush()

        # Get or create session without starting heartbeat
        session = self.context.sockjs_server.get_session(session_id)
        if session is None:
            session = self.context.sockjs_server.create_session(session_id)
        if session is None:
            # close the session in the next cycle.
            self.context.gaplic.add_callback(self.session_closed)
            return response  # how inform of the error? headers has been sent.

        # Try to attach to the session
        if not session.set_handler(self, False):
            # close the session in the next cycle.
            self.context.gaplic.add_callback(self.session_closed)
            return response  # how inform of the error? headers has been sent.
        self.session = session
        # Verify if session is properly opened
        session.verify_state()

        session.flush()

        return response

    def send_pack(self, message, binary=False):
        if binary:
            raise Exception(
                'binary not supported for EventsourceStreamingTransport'
            )

        msg = 'data: %s\r\n\r\n' % message

        #self.active = False  # TODO si chequeo active fallan los tests

        try:
            self.notify_sent(len(msg))
            self.write(msg)
            self.flush(callback=self.send_complete)
        except Exception:
            logging.exception(
                "ERROR EventsourceStreamingTransport send_pack")
            # If connection dropped, make sure we close offending session
            # instead of propagating error all way up.
            if self.session:  # detach session
                self.session.delayed_close()
                self.session.remove_handler(self)
                self.session = None

    def send_complete(self):
        """
            Verify if connection should be closed based on amount of data
            that was sent.
        """
        self.active = True

        if self.should_finish():
            if self.session:  # detach session
                self.session.remove_handler(self)
                self.session = None
            self.context.gaplic.add_callback(self.safe_finish)
        else:
            if self.session:
                self.session.flush()

    def notify_sent(self, data_len):
        """
            Update amount of data sent
        """
        self.amount_limit -= data_len

    def should_finish(self):
        """
            Check if transport should close long running connection after
            sending X bytes to the client.

            `data_len`
                Amount of data that was sent
        """
        if self.amount_limit <= 0:
            return True

        return False

    def session_closed(self):
        """Called by the session when it was closed"""
        if self.session:  # detach session
            self.session.remove_handler(self)
            self.session = None
        self.context.gaplic.add_callback(self.safe_finish)



# -*- encoding: utf-8 -*-
""" GSockjsServer GObj

.. autoclass:: GSockjsServer
    :members: start_up

"""

from pyramid.security import (
    Allow,
    Everyone,
    DENY_ALL,
)

from ginsfsm.gobj import GObj
from ginsfsm.protocols.sockjs.server.c_static import (
    GInfo,
    GIFrame,
)
from ginsfsm.c_timer import GTimer
from ginsfsm.deferred import Deferred
from ginsfsm.protocols.sockjs.server.c_transport_websocket import GRawWebsocket
from ginsfsm.protocols.sockjs.server import session, sessioncontainer
from ginsfsm.protocols.sockjs.server import stats
from ginsfsm.protocols.sockjs.server.c_session import GSockjsSession
from ginsfsm.protocols.sockjs.server import proto


#-------------------------------------------#
#               Actions
#-------------------------------------------#
def ac_timeout(self, event):
    self.counter += 1
    self.send_event(self.timer, 'EV_SET_TIMER', seconds=self.seconds)


def ac_sample(self, event):
    """ Event attributes:
        * :attr:`data`: example.

    """
    self.broadcast_event('EV_OUTPUT_EVENT', data=event.data)


GSOCKJSSERVER_FSM = {
    'event_list': (
        'EV_INPUT_EVENT:top input',
        'EV_OUTPUT_EVENT:top output',
    ),
    'state_list': (
        'ST_IDLE',
    ),
    'machine': {
        'ST_IDLE':
        (
            ('EV_INPUT_EVENT',      ac_sample,      None),
        ),
    }
}

GSOCKJSSERVER_GCONFIG = {
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
        "Default is ``None``, i.e., the parent"
    ],
    'sockjs_app_class': [
        None, None, 0, None,
        "Class to receive sockjs events."
        "GSockjsServer will create a instance of this class."
    ],

    # like tornado settings
    'session_check_interval': [
        int, 5, 0, None, "Sessions check interval in seconds"
    ],
    'disconnect_delay': [
        int, 10, 0, None, "Session expiration in seconds"
    ],
    'heartbeat_delay': [
        int, 25, 0, None,
        "Heartbeat time in seconds. Do not change this value unless "
        "you absolutely sure that new value will work."
    ],
    'disabled_transports': [tuple, (), 0, None, "Transports disabled"],
    'sockjs_url': [
        str, 'http://cdn.sockjs.org/sockjs-0.3.min.js', 0, None,
        "SockJS location"
    ],
    'response_limit': [int, 128 * 1024, 0, None, "Max response body size"],
    'jsessionid': [
        bool, True, 0, None,
        "Enable or disable JSESSIONID cookie handling"
    ],
    'immediate_flush': [
        bool, True, 0, None,
        "Should sockjs-tornado flush messages immediately or queue then and"
        " flush on next ioloop tick"
    ],
    'verify_ip': [
        bool, True, 0, None,
        "Enable IP checks for polling transports. If enabled, "
        "all subsequent polling calls should be from the same IP address."
    ],
    'per_user': [
        bool, False, 0, None,
        "TODO: default is True in pyramid_sockjs"
    ],
}


class GSockjsServer(GObj):
    """  GSockjsServer GObj.
    This gobj only works
    with `Pyramid <http://www.pylonsproject.org/>`_ framework.

    .. ginsfsm::
       :fsm: GSOCKJSSERVER_FSM
       :gconfig: GSOCKJSSERVER_GCONFIG

    *Input-Events:*
        * :attr:`'EV_INPUT_EVENT'`: sample input event.

          Event attributes:

              * ``data``: sample event attribute.

    *Output-Events:*
        * :attr:`'EV_OUTPUT_EVENT'`: sample output event.

          Event attributes:

              * ``data``: sample event attribute.

    """
    __acl__ = [
        (Allow, Everyone, 'view'),
        (DENY_ALL),
    ]

    def __init__(self):
        GObj.__init__(self, GSOCKJSSERVER_FSM, GSOCKJSSERVER_GCONFIG)
        # Store connection class

    def start_up(self):
        """ Initialization zone.
        """
        self.logger.info(
            "==> Resource path of sockjs server: %s\n"
            "==> Test connection with => %s%s/%s" % (
            self.resource_path(),
            'schema://host:port',  # TODO: get real values
            self.resource_path(),
            'info'))
        if self.config.subscriber is None:
            self.config.subscriber = self.parent
        self.subscribe_event(None, self.config.subscriber)

        if not self.config.sockjs_app_class:
            raise Exception('You must supply a connection application.')
        self._connection = self.config.sockjs_app_class

        #----------------------------------------------#
        #           Static handlers
        #----------------------------------------------#

        # Info
        self.create_gobj(
            'info',
            GInfo,
            self,
            sockjs_server=self,
        )
        # IFrame
        self.create_gobj(
            'iframe',
            GIFrame,
            self,
            re_name='iframe[0-9-.a-z_]*.html',
            sockjs_server=self,
        )
        # RawWebsocket
        self.create_gobj(
            'websocket',
            GRawWebsocket,
            self,
            sockjs_server=self,
        )

        #----------------------------------------------#
        #           Session URLs
        #----------------------------------------------#
        self.sockjs_session = self.create_gobj(
            'sockjs_session',
            GSockjsSession,
            self,
            re_name='^[^/.]+$',
            sockjs_server=self,
        )

        #----------------------------------------------#
        #           Setup
        #----------------------------------------------#
        # Store connection class
        #self._connection = connection

        # Initialize io_loop
        #self.io_loop = io_loop or ioloop.IOLoop.instance()

        disabled_transports = self.config.disabled_transports
        self.websockets_enabled = 'websocket' not in disabled_transports
        self.cookie_needed = self.config.jsessionid

        # Sessions
        self._sessions = sessioncontainer.SessionContainer()

        check_interval = self.config.session_check_interval

        self.timer = self.create_gobj(
            None,       # unnamed gobj
            GTimer,     # gclass
            self        # parent
        )
        self.timer.delete_all_subscriptions()
        deferred = Deferred(None, self._sessions.expire)
        self.timer.subscribe_event(
            'EV_TIMEOUT',
            deferred,
            __deferred_witout_oevent__=True,
        )
        self.send_event(
            self.timer,
            'EV_SET_TIMER',
            seconds=check_interval,
            autostart=True)

        #self._sessions_cleanup = PeriodicCallback(
        #    self._sessions.expire,
        #    check_interval,
        #    self.gaplic)
        #self._sessions_cleanup.start()

        # Stats
        self.stats = stats.StatsCollector()

    def create_session(self, session_id, register=True):
        """Creates new session object and returns it.

        `request`
            Request that created the session. Will be used to get query string
            parameters and cookies
        `register`
            Should be session registered in a storage. Websockets don't
            need it.
        """
        # TODO: Possible optimization here for settings.get
        s = session.Session(
            self._connection,
            self,
            session_id,
            self.config.disconnect_delay
        )
        self.logger.debug("CREATE SESSION: id=%s, s=%r" % (session_id, s,))
        if register:
            self._sessions.add(s)

        return s

    def get_session(self, session_id):
        """Get session by session id

        `session_id`
            Session id
        """
        s = self._sessions.get(session_id)
        self.logger.debug("GET SESSION: id=%s, s=%r" % (session_id, s,))
        return s

    def get_connection_class(self):
        """Return associated connection class"""
        return self._connection

    # Broadcast helper
    def broadcast(self, clients, msg):
        """Optimized `broadcast` implementation.
        Depending on type of the session, will json-encode
        message once and will call either `send_message` or `send_jsonifed`.

        `clients`
            Clients iterable
        `msg`
            Message to send
        """
        json_msg = None

        count = 0

        for c in clients:
            sess = c.session
            if not sess.is_closed:
                if sess.send_expects_json:
                    if json_msg is None:
                        json_msg = proto.json_encode(msg)
                    sess.send_jsonified(json_msg, False)
                else:
                    sess.send_message(msg, stats=False)

                count += 1

        self.stats.on_pack_sent(count)


#----------------------------------------------------------------#
#                   Views
#----------------------------------------------------------------#
from ginsfsm.protocols.sockjs.server.basehandler import BaseHandler
from pyramid.view import view_config


@view_config(
    context=GSockjsServer,
    name='',
    attr='get',
    request_method='GET',
)
class GreetingsHandler(BaseHandler):
    """SockJS greetings page handler"""
    def get(self):
        response = self.response
        self.enable_cache()
        response.content_type = 'text/plain'
        response.charset = 'UTF-8'
        data = 'Welcome to SockJS!\n'.encode()
        response.body = data
        return response



# -*- encoding: utf-8 -*-
""" SockJSConnection GObj

.. autoclass:: SockJSConnection
    :members: start_up

"""

import math

from ginsfsm.protocols.sockjs.server.conn import SockJSConnection


class EchoConnection(SockJSConnection):
    def on_message(self, msg):
        self.send(msg)


class CloseConnection(SockJSConnection):
    def on_open(self, info):
        self.close()

    def on_message(self, msg):
        pass


class CookieEcho(SockJSConnection):
    def on_message(self, msg):
        self.send(msg)



# -*- encoding: utf-8 -*-
"""
GObj :class:`GServerSock`
=========================

GObj for manage server socket.

.. autoclass:: GServerSock
    :members:

"""
try:
    import ssl  # Python 2.6+
except ImportError:
    ssl = None

import errno
import socket
from ginsfsm.gobj import GObj
from ginsfsm.c_sock import (
    GSock,
    GSSLSock
)


def ac_disconnected(self, event):  # ONLY for documentation (It's in GSock)
    """ Used by GServerSock.
    """
    gsock = event.gsock
    self.post_event(self, 'EV_DESTROY_GSOCK', gsock=gsock)


def ac_drop_gsock(self, event):  # ONLY for documentation (It's in GSock)
    """ Used by GServerSock.
        Destroy gsock.
    """
    gsock = event.gsock
    GObj.destroy_gobj(gsock)
    self.config.n_gsocks -= 1


GSERVERSOCK_FSM = {  # ONLY for documentation. Partial use of GSOCK_FSM.
    'event_list': (
        'EV_CONNECTED: top output',
        'EV_DISCONNECTED: bottom input',
        'EV_DESTROY_GSOCK',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_DISCONNECTED',     ac_disconnected,    None),
            ('EV_DESTROY_GSOCK',    ac_drop_gsock,      None),
        ),
    }
}

GSERVERSOCK_GCONFIG = {  # ONLY for documentation. Partial use of GSOCK_CONFIG.
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
        "Default is ``None``, i.e., the parent"
    ],
    'host': [str, '', 0, None, "server host (ip or name)"],
    'port': [int, 0, 0, None, "server port"],
    'ports': [tuple, (), 0, None, "multi ports"],
    'use_multi_ports': [bool, False, 0, None, "use multi ports"],
    'ssl_options': [dict, {}, 0, None, "ssl options"],
    'n_gsocks': [int, 0, 0, None, "Server Stats: Number of gsocks opened"],
}


class GServerSock(GSock):
    """  GSock gobj acting as server sock.

    This gobj is derived from :class:`ginsfsm.c_sock.GSock`,
    but it uses the server behaviour of a sock.
    It creates a listening sock at configured (host, port) attributes.

    `GServerSock` can serve SSL traffic with Python 2.6+ and OpenSSL.
    To make this server serve SSL traffic, send the ssl_options dictionary
    argument with the arguments required for the `ssl.wrap_socket` method,
    including "certfile" and "keyfile"::

       ssl_options={
           "certfile": os.path.join(data_dir, "mydomain.crt"),
           "keyfile": os.path.join(data_dir, "mydomain.key"),
       }

    Each incoming connection
    will create a new :class:`ginsfsm.c_sock.GSock` :term:`gobj`,
    that it will be child of the :attr:`subscriber` `gobj`
    (the parent by default).

        .. note:: Once an ``'EV_CONNECTED'`` event has been sent,
           the GServerSock
           doesn't know anything more of the new gsock client.

           The relationship is directly between the
           accepted :class:`ginsfsm.c_sock.GSock` gobj
           and the :attr:`subscriber`.

           The `subcriber` is responsible of receiving all events
           and destroy the new gsock client.


        .. warning::  Remember to destroy the accepted `gobj`
           with :func:`destroy_gobj` when the `gobj` has been disconnected.

           The `subcriber` knows when a new `gobj` has been accepted because it
           receives the ``'EV_CONNECTED'`` event.

           When the `subcriber` receives a ``'EV_DISCONNECTED'`` event must
           destroy the `gobj` because the connection ceases to exist.

    .. ginsfsm::
       :fsm: GSERVERSOCK_FSM
       :gconfig: GSERVERSOCK_GCONFIG

    *Output-Events:*
        * :attr:`'EV_CONNECTED'`: new client socket connected.

          The gobj sending the `'EV_CONNECTED'` event
          is the new client socket gobj.

          Attributes added to the sent :term:`event`:

            * ``peername``: remote address to which the socket is connected.
            * ``sockname``: the socket’s own address.

          The :attr:`subscriber` can know from which server comes the client
          by the name of client gobj.

          If the server gobj has name, the name of client gobj will be:
          'server-name.clisrv_#'
    """

    def __init__(self):
        GSock.__init__(self)
        self._n_clisrv = 0

    def start_up(self):
        """ Initialization zone.

        Subscribe all events to parent if :attr:`subscriber` is None.

        Start listen to (host,port).
        """
        if self.config.subscriber is None:
            self.config.subscriber = self.parent
        self.do_listen()

    def do_listen(self):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        host = self.config.host
        port = 0
        if not self.config.use_multi_ports:
            port = int(self.config.port)
            self.bind((host, port))
        else:
            self.port = 0
            ports = self.config.ports
            for port in ports:
                try:
                    port = int(port)
                    self.bind((host, port))
                except socket.error:
                    port = 0
                    continue
                else:
                    self.port = port
                    break
        if port:
            if self.logger:
                self.logger.info("Listening %r at host %r, port %r..." % (
                    self, host, port)
                )
            self.listen(1024)
        else:
            self.logger.error("ERROR at %r NO listening port!", self)

    def handle_accepted(self, sock, addr):
        # The socket options to set on receiving a connection.
        # It is a list of (level, optname, value) tuples.
        # TCP_NODELAY disables the Nagle algorithm for writes
        # (Waitress already buffers its writes).
        # TODO: check origins for permitted source ip.
        socket_options = [
            (socket.SOL_TCP, socket.TCP_NODELAY, 1),
        ]
        for (level, optname, value) in socket_options:
            sock.setsockopt(level, optname, value)

        # copied from tornado.netutil.py
        if self.config.ssl_options:
            assert ssl, "Python 2.6+ and OpenSSL required for SSL"
            try:
                sock = ssl.wrap_socket(
                    sock,
                    server_side=True,
                    do_handshake_on_connect=False,
                    **self.config.ssl_options
                )
            except ssl.SSLError as err:
                if err.args[0] == ssl.SSL_ERROR_EOF:
                    return sock.close()
                else:
                    raise
            except socket.error as err:
                if err.args[0] == errno.ECONNABORTED:
                    return sock.close()
                else:
                    raise

        if self.config.ssl_options:
            gsock_class = GSSLSock
        else:
            gsock_class = GSock

        self._n_clisrv += 1
        clisrv = self.create_gobj(
            'clisrv_%x' % self._n_clisrv,
            gsock_class,
            self,
            subscriber=self.config.subscriber,
        )
        clisrv.set_clisrv_socket(sock)
        clisrv.handle_connect()

        # we need to know disconnected event for deleting gobjs
        clisrv.subscribe_event(
            'EV_DISCONNECTED',
            self,
            __hard_subscription__=True,
        )
        self.config.n_gsocks += 1



import unittest
from ginsfsm.circular_fifo import (
    CircularFIFO,
    CircularFullBufferError,
)


class TestCircular(unittest.TestCase):
    def setUp(self):
        self.circular = CircularFIFO(5)

    def test_getdata_putdata_without_overwrite(self):
        data = self.circular.getdata(1)
        self.assertEqual(data, b'')

        count = self.circular.free_space
        self.assertEqual(count, 5)

        count = self.circular.putdata(b"123")
        self.assertEqual(count, 3)
        count = self.circular.free_space
        self.assertEqual(count, 2)

        count = self.circular.putdata(b"ab")
        self.assertEqual(count, 2)
        count = self.circular.free_space
        self.assertEqual(count, 0)

        self.assertRaises(CircularFullBufferError,
            self.circular.putdata,
            b"XX")

        data = self.circular.getdata(5)
        self.assertEqual(data, b"123ab")

        data = self.circular.getdata(1)
        self.assertEqual(data, b'')

        count = self.circular.putdata(b"qw")
        self.assertEqual(count, 2)
        data = self.circular.getdata()
        self.assertEqual(data, b'qw')

    def test_getdata_putdata_with_overwrite(self):
        count = self.circular.putdata(b"123")
        self.assertEqual(count, 3)
        count = self.circular.free_space
        self.assertEqual(count, 2)

        self.assertRaises(CircularFullBufferError,
            self.circular.putdata,
            b"abc")

        data = self.circular.getdata(2)
        self.assertEqual(data, b"12")
        count = self.circular.free_space
        self.assertEqual(count, 4)
        count = self.circular.busy_space
        self.assertEqual(count, 1)

        count = self.circular.putdata(b"abcd")
        self.assertEqual(count, 4)

        data = self.circular.getdata(0)
        self.assertEqual(data, b"3abcd")
        count = self.circular.free_space
        self.assertEqual(count, 5)

        count = self.circular.putdata(b"zxcvb")
        self.assertEqual(count, 5)
        count = self.circular.free_space
        self.assertEqual(count, 0)

        data = self.circular.getdata(4)
        self.assertEqual(data, b"zxcv")
        count = self.circular.free_space
        self.assertEqual(count, 4)

        data = self.circular.getdata(3)
        self.assertEqual(data, b"b")
        count = self.circular.free_space
        self.assertEqual(count, 5)

    def test_getdata_putdata_with_ln(self):
        s = bytearray(b'123')
        count = self.circular.putdata(s, 2)
        self.assertEqual(count, 2)
        count = self.circular.free_space
        self.assertEqual(count, 3)

        s = b"abc"
        count = self.circular.putdata(s, 5)
        self.assertEqual(count, 3)
        count = self.circular.free_space
        self.assertEqual(count, 0)

        data = self.circular.getdata()
        self.assertEqual(data, b"12abc")

if __name__ == "__main__":
    circular = CircularFIFO(5)

    circular.putdata(b"123")
    print(circular)
    circular.getdata(2)
    print(circular)

    circular.putdata(b"abcd")
    print(circular)

    circular.getdata(0)
    print(circular)



# -*- encoding: utf-8 -*-
"""
A python library to do development based in Finite State Machines.
"""
__version__ = '0.6.8'
__title__ = 'ginsfsm'
__author__ = 'Ginés Martínez'
__license__ = 'MIT License'
__copyright__ = 'Copyright 2012-2013 Ginés Martínez'




# -*- encoding: utf-8 -*-
"""
bolsa
=====

It uses :class:`ginsfsm.protocols.wsgi.server.c_wsgi_server.GWsgiServer`.

.. autofunction:: main

"""

from ginsfsm.gaplic import GAplic
from ginsfsm.globals import (
    set_global_main_gaplic,
)
from bolsa.c_acciones import GAcciones


#===============================================================
#                   Main
#       Paste app factory.
#       To run with gserve bolsa.ini
#===============================================================
def main(global_config, **local_conf):
    """ Entry point to run with gserve (PasteDeploy)
    """
    if 'gaplic-name' in local_conf:
        gaplic_name = local_conf.pop('gaplic-name')
    else:
        raise Exception('You must supply an gaplic name ("gaplic-name")')

    ga = GAplic(
        name=gaplic_name,
        roles=('bolsa',),
        **local_conf)
    set_global_main_gaplic(ga)

    ga.create_gobj(
        'acciones',
        GAcciones,
        ga,
        __unique_name__=True,
    )
    return ga


if __name__ == "__main__":
    """ You can run directly this file, without gserve.
    """
    import logging
    logging.basicConfig(level=logging.DEBUG)

    # simulate running from ini file
    local_conf = {
        'gaplic-name': 'bolsa',
        'router_enabled':  True,
        'GRouter.server':  True,
        'GRouter.localhost_route_ports':  8002,
        'GRouter.trace_router':  True,
        'GObj.trace_mach':  True,
        'GObj.trace_creation':  False,
        'GObj.trace_traverse':  True,
        'GObj.trace_subscription':  False,
        'GSock.trace_dump':  True,
    }
    ga = main({}, **local_conf)

    try:
        ga.start()
    except (KeyboardInterrupt, SystemExit):
        print('Program stopped')



from ginsfsm.smachine import SMachine
from .fsmdraw import FsmDraw


def fsm2image(filepath, fsm, fsm_name, output_format,
              maxwidth=None,
              font_name='./DejaVuSans.ttf',
              font_size=14):
    """ Convert the fsm into a sphinx image.
    """
    smachine = SMachine(fsm)
    smachine.name = fsm_name

    # generate the svg file
    drawer = FsmDraw(smachine, filepath, font_name, font_size)
    drawer.draw()
    drawer.save()

    if output_format == 'png':
        # todo
        pass
    elif output_format == 'pdf':
        # todo
        pass

    """
    size = drawer.pagesize()
    if maxwidth and maxwidth < size[0]:
        ratio = float(maxwidth) / size[0]
        thumb_size = (maxwidth, size[1] * ratio)

        thumb_filepath = '_thumb' + filepath
        drawer.filename = thumb_filepath
        drawer.draw()
        drawer.save(thumb_size)

        image = nodes.image(uri=thumb_filepath, target=filepath)
    else:
        image = nodes.image(uri='/_static/' + filename + '.svg')
        #from ginsfsm.sphinxext import xxxmath
        #image = xxxmath(uri='../../images/' + filename + '.svg')
    """



# (c) 2005 Ian Bicking and contributors;
# written for Paste (http://pythonpaste.org)
# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
"""
Map URL prefixes to WSGI applications.  See ``URLMap``
"""

from ginsfsm.compat import escape
import re

__all__ = ['URLMap']

class DictMixin:
    # Mixin defining all dictionary methods for classes that already have
    # a minimum dictionary interface including getitem, setitem, delitem,
    # and keys. Without knowledge of the subclass constructor, the mixin
    # does not define __init__() or copy().  In addition to the four base
    # methods, progressively more efficiency comes with defining
    # __contains__(), __iter__(), and iteritems().

    # second level definitions support higher levels
    def __iter__(self):
        for k in self.keys():
            yield k
    def has_key(self, key):
        try:
            self[key]
        except KeyError:
            return False
        return True
    def __contains__(self, key):
        return self.has_key(key)

    # third level takes advantage of second level definitions
    def iteritems(self):
        for k in self:
            yield (k, self[k])
    def iterkeys(self):
        return self.__iter__()

    # fourth level uses definitions from lower levels
    def itervalues(self):
        for _, v in self.iteritems():
            yield v
    def values(self):
        return [v for _, v in self.iteritems()]
    def items(self):
        return list(self.iteritems())
    def clear(self):
        for key in self.keys():
            del self[key]
    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default
    def pop(self, key, *args):
        if len(args) > 1:
            raise TypeError("pop expected at most 2 arguments, got "\
                              + repr(1 + len(args)))
        try:
            value = self[key]
        except KeyError:
            if args:
                return args[0]
            raise
        del self[key]
        return value
    def popitem(self):
        try:
            k, v = self.iteritems().next()
        except StopIteration:
            raise KeyError('container is empty')
        del self[k]
        return (k, v)
    def update(self, other=None, **kwargs):
        # Make progressively weaker assumptions about "other"
        if other is None:
            pass
        elif hasattr(other, 'iteritems'):  # iteritems saves memory and lookups
            for k, v in other.iteritems():
                self[k] = v
        elif hasattr(other, 'keys'):
            for k in other.keys():
                self[k] = other[k]
        else:
            for k, v in other:
                self[k] = v
        if kwargs:
            self.update(kwargs)
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    def __repr__(self):
        return repr(dict(self.iteritems()))
    def __cmp__(self, other):
        if other is None:
            return 1
        if isinstance(other, DictMixin):
            other = dict(other.iteritems())
        return cmp(dict(self.iteritems()), other)
    def __len__(self):
        return len(self.keys())


def urlmap_factory(loader, global_conf, **local_conf):
    """
        # TODO: check
        [composite:urlmap]
        use = call:ginsfsm.protocols.wsgi.common.urlmap:urlmap_factory
        / = home
        /blog = blog
        /wiki = wiki

        [app:home]
        use = call:ginsfsm.examples.wsgi.simple_wsgi_server:paste_app_factory

        [app:blog]
        use = call:ginsfsm.examples.wsgi.simple_wsgi_server:paste_app_factory

        [app:wiki]
        use = call:ginsfsm.examples.wsgi.simple_wsgi_server:paste_app_factory
    """
    if 'not_found_app' in local_conf:
        not_found_app = local_conf.pop('not_found_app')
    else:
        not_found_app = global_conf.get('not_found_app')
    if not_found_app:
        not_found_app = loader.get_app(not_found_app, global_conf=global_conf)
    urlmap = URLMap(not_found_app=not_found_app)
    for path, app_name in local_conf.items():
        path = parse_path_expression(path)
        app = loader.get_app(app_name, global_conf=global_conf)
        urlmap[path] = app
    return urlmap


def parse_path_expression(path):
    """
    Parses a path expression like 'domain foobar.com port 20 /' or
    just '/foobar' for a path alone.  Returns as an address that
    URLMap likes.
    """
    parts = path.split()
    domain = port = path = None
    while parts:
        if parts[0] == 'domain':
            parts.pop(0)
            if not parts:
                raise ValueError(
                    "'domain' must be followed with a domain name")
            if domain:
                raise ValueError("'domain' given twice")
            domain = parts.pop(0)
        elif parts[0] == 'port':
            parts.pop(0)
            if not parts:
                raise ValueError("'port' must be followed with a port number")
            if port:
                raise ValueError("'port' given twice")
            port = parts.pop(0)
        else:
            if path:
                raise ValueError("more than one path given (have %r, got %r)"
                                 % (path, parts[0]))
            path = parts.pop(0)
    s = ''
    if domain:
        s = 'http://%s' % domain
    if port:
        if not domain:
            raise ValueError("If you give a port, you must also give a domain")
        s += ':' + port
    if path:
        if s:
            s += '/'
        s += path
    return s


class URLMap(DictMixin):

    """
    URLMap instances are dictionary-like object that dispatch to one
    of several applications based on the URL.

    The dictionary keys are URLs to match (like
    ``PATH_INFO.startswith(url)``), and the values are applications to
    dispatch to.  URLs are matched most-specific-first, i.e., longest
    URL first.  The ``SCRIPT_NAME`` and ``PATH_INFO`` environmental
    variables are adjusted to indicate the new context.

    URLs can also include domains, like ``http://blah.com/foo``, or as
    tuples ``('blah.com', '/foo')``.  This will match domain names; without
    the ``http://domain`` or with a domain of ``None`` any domain will be
    matched (so long as no other explicit domain matches).  """

    def __init__(self, not_found_app=None):
        self.applications = []
        if not not_found_app:
            not_found_app = self.not_found_app
        self.not_found_application = not_found_app

    norm_url_re = re.compile('//+')
    domain_url_re = re.compile('^(http|https)://')

    def not_found_app(self, environ, start_response):
        # TODO: must change to ginsfsm http response!
        mapper = environ.get('paste.urlmap_object')
        if mapper:
            matches = [p for p, a in mapper.applications]
            extra = 'defined apps: %s' % (
                ',\n  '.join(list(map(repr, matches))))
        else:
            extra = ''
        extra += '\nSCRIPT_NAME: %r' % environ.get('SCRIPT_NAME')
        extra += '\nPATH_INFO: %r' % environ.get('PATH_INFO')
        extra += '\nHTTP_HOST: %r' % environ.get('HTTP_HOST')
        # NotFound('The resource could not be found.')
        # TODO: eliminate paste dependency
        from paste import httpexceptions
        app = httpexceptions.HTTPNotFound(
            environ['PATH_INFO'],
            comment=escape(extra)).wsgi_application
        return app(environ, start_response)

    def normalize_url(self, url, trim=True):
        if isinstance(url, (list, tuple)):
            domain = url[0]
            url = self.normalize_url(url[1])[1]
            return domain, url
        assert (not url or url.startswith('/')
                or self.domain_url_re.search(url)), (
            "URL fragments must start with / or http:// (you gave %r)" % url)
        match = self.domain_url_re.search(url)
        if match:
            url = url[match.end():]
            if '/' in url:
                domain, url = url.split('/', 1)
                url = '/' + url
            else:
                domain, url = url, ''
        else:
            domain = None
        url = self.norm_url_re.sub('/', url)
        if trim:
            url = url.rstrip('/')
        return domain, url

    def sort_apps(self):
        """
        Make sure applications are sorted with longest URLs first
        """
        def key(app_desc):
            (domain, url), app = app_desc
            if not domain:
                # Make sure empty domains sort last:
                return '\xff', -len(url)
            else:
                return domain, -len(url)
        apps = [(key(desc), desc) for desc in self.applications]
        apps.sort()
        self.applications = [desc for (sortable, desc) in apps]

    def __setitem__(self, url, app):
        if app is None:
            try:
                del self[url]
            except KeyError:
                pass
            return
        dom_url = self.normalize_url(url)
        if dom_url in self:
            del self[dom_url]
        self.applications.append((dom_url, app))
        self.sort_apps()

    def __getitem__(self, url):
        dom_url = self.normalize_url(url)
        for app_url, app in self.applications:
            if app_url == dom_url:
                return app
        raise KeyError(
            "No application with the url %r (domain: %r; existing: %s)"
            % (url[1], url[0] or '*', self.applications))

    def __delitem__(self, url):
        url = self.normalize_url(url)
        for app_url, app in self.applications:
            if app_url == url:
                self.applications.remove((app_url, app))
                break
        else:
            raise KeyError(
                "No application with the url %r" % (url,))

    def keys(self):
        return [app_url for app_url, app in self.applications]

    def __call__(self, environ, start_response):
        host = environ.get('HTTP_HOST', environ.get('SERVER_NAME')).lower()
        if ':' in host:
            host, port = host.split(':', 1)
        else:
            if environ['wsgi.url_scheme'] == 'http':
                port = '80'
            else:
                port = '443'
        path_info = environ.get('PATH_INFO')
        path_info = self.normalize_url(path_info, False)[1]
        for (domain, app_url), app in self.applications:
            if domain and domain != host and domain != host + ':' + port:
                continue
            if (path_info == app_url
                or path_info.startswith(app_url + '/')):
                environ['SCRIPT_NAME'] += app_url
                environ['PATH_INFO'] = path_info[len(app_url):]
                return app(environ, start_response)
        environ['paste.urlmap_object'] = self
        return self.not_found_application(environ, start_response)



# -*- encoding: utf-8 -*-
"""
GObj :class:`GHttpServer`
=========================

.. autoclass:: GHttpServer
    :members:

"""
try:
    import ssl  # Python 2.6+
except ImportError:
    ssl = None

from ginsfsm.gobj import GObj
from ginsfsm.c_srv_sock import GServerSock
from ginsfsm.protocols.http.server.c_http_clisrv import GHttpCliSrv


def ac_connected(self, event):
    """ New clisvr gsock (http channel):
    A new client http has been acepted.
    The new clisrv GSock, created by GServerSock is sending this event.
    """
    gsock = event.source[-1]
    gsock.delete_all_subscriptions()

    self._n_channel += 1
    clisrv = self.create_gobj(
        'channel_%x' % self._n_channel,
        GHttpCliSrv,
        self,
        gsock=gsock,
        subscriber=self.config.subscriber,  # TODO better to join two configs
        maximum_simultaneous_requests=self.config.maximum_simultaneous_requests,
        url_scheme=self.config.url_scheme,
        inactivity_timeout=self.config.inactivity_timeout,
        responseless_timeout=self.config.responseless_timeout,
        inbuf_overflow=self.config.inbuf_overflow,
        max_request_header_size=self.config.max_request_header_size,
        max_request_body_size=self.config.max_request_body_size,
    )
    clisrv.subscribe_event(  # to delete the channel
        'EV_HTTP_CHANNEL_CLOSED',
        self,
        __hard_subscription__=True,
    )
    self.config.n_channels += 1


def ac_channel_closed(self, event):
    """ **Some** clisvr gsock closed, drop it.
    """
    channel = event.channel
    self.post_event(self, 'EV_DESTROY_CHANNEL', channel=channel)


def ac_drop_httpchannel(self, event):
    """ It's better receive this event by post_event().
    """
    channel = event.channel
    self._n_channel -= 1
    # the gsock is destroy in channel.ac_disconnected (GHttpCliSrv)
    GObj.destroy_gobj(channel)
    self.config.n_channels -= 1


def ac_timeout(self, event):
    self.set_timeout(10)
    #print("Server's clients: %d, connected %d" % (
    #    len(self.dl_childs), self._n_channel))


GHTTPSERVER_FSM = {
    'event_list': (
        'EV_CONNECTED: bottom input',
        'EV_HTTP_CHANNEL_CLOSED: bottom input',
        'EV_TIMEOUT',
        'EV_DESTROY_CHANNEL',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_TIMEOUT',              ac_timeout,             None),
            ('EV_CONNECTED',            ac_connected,           None),
            ('EV_HTTP_CHANNEL_CLOSED',  ac_channel_closed,      None),
            ('EV_DESTROY_CHANNEL',      ac_drop_httpchannel,    None),
        ),
    }
}

GHTTPSERVER_GCONFIG = {
    'subscriber': [
        None, None, 0, None,
        "subcriber of all output-events."
        "Default is ``None``, i.e., the parent"
    ],
    'host': [str, '', 0, None, "listening host"],
    'port': [int, 0, 0, None, "listening port"],
    'origins': [
        None, None, 0, None,
        "TODO:list of (host, port) tuples allowed to connect from"
    ],

    'expose_tracebacks': [
        bool, False, 0, None, "expose tracebacks of uncaught exceptions"
    ],
    'maximum_simultaneous_requests': [
        int, 0, 0, None,
        "maximum simultaneous requests. Default: 0, without limit."
    ],
    'identity': [
        str, 'ginsfsm', 0, None,
        "server identity (sent in Server: header)"
    ],
    'url_scheme': [str, 'http', 0, None, "default ``http`` value"],
    'inactivity_timeout': [
        int, 5 * 60 * 60, 0, None,
        "Inactivity timeout in seconds."
    ],
    'responseless_timeout': [
        int, 5 * 60 * 60, 0, None,
        "'Without response' timeout in seconds."
    ],
    # A tempfile should be created if the pending input is larger than
    # inbuf_overflow, which is measured in bytes. The default is 512K.  This
    # is conservative.
    'inbuf_overflow': [int, 524288, 0, None, ""],
    # maximum number of bytes of all request headers combined (256K default)
    'max_request_header_size': [int, 262144, 0, None, ""],
    # maximum number of bytes in request body (1GB default)
    'max_request_body_size': [int, 1073741824, 0, None, ""],
    'n_channels': [int, 0, 0, None, "Server stats, number of channels opened"],
}


class GHttpServer(GObj):
    """  Asynchronous Http Server gobj.

    In the startup, this class creates internally a server socket, gobj of
    :class:`ginsfsm.c_srv_sock.GServerSock` gclass.

    When a new client connects,
    a new :class:`ginsfsm.c_sock.GSock` gobj is created,
    receiving the EV_CONNECTED event.
    Then it creates a
    :class:`ginsfsm.protocols.http.server.c_http_clisrv.GHttpCliSrv` gobj
    that will process all the events of the GSock gobj.

    The output events of
    :class:`ginsfsm.protocols.http.server.c_http_clisrv.GHttpCliSrv`
    will be processed by ``subscriber`` gobj of this class.

    .. ginsfsm::
       :fsm: GHTTPSERVER_FSM
       :gconfig: GHTTPSERVER_GCONFIG

    *Input-Events:*
        * :attr:`'EV_CONNECTED'`: new *clisrv*, client socket.

          The internal child :class:`ginsfsm.c_srv_sock.GServerSock` gobj
          has accepted a new socket connection,
          and it has created a new client gobj (*clisrv*) of
          :class:`ginsfsm.c_sock.GSock` gclass.

          This event has been sended
          by the new :class:`ginsfsm.c_sock.GSock` *clisrv* gobj.

          A new
          :class:`ginsfsm.protocols.http.server.c_http_clisrv.GHttpCliSrv` gobj
          is created, to process all the events of the GSock gobj.

          This class will be subscred to the ``'EV_HTTP_CHANNEL_CLOSED'`` event
          in order to destroy the GHttpCliSrv gobj.

          Event attributes:

            * ``peername``: remote address to which the socket is connected.
            * ``sockname``: the socket’s own address.


        * :attr:`'EV_HTTP_CHANNEL_CLOSED'`: http channel disconnected.

          The http server subcribes this event from clisrv gobj,
          in order to destroy it when became disconnected.
    """

    def __init__(self):
        GObj.__init__(self, GHTTPSERVER_FSM, GHTTPSERVER_GCONFIG)
        self._n_channel = 0

    def start_up(self):
        self.gserversock = self.create_gobj(
            self.name if self.name else 'sock-server',
            GServerSock,
            self,
            subscriber=self,  # Iniatially capture all events from new clisrv.
            host=self.config.host,
            port=self.config.port,
        )

    def _get_ssl_certificate(self, binary_form=False):
        """Returns the client's SSL certificate, if any.

        To use client certificates, the HTTPServer must have been constructed
        with cert_reqs set in ssl_options, e.g.::

            server = HTTPServer(app,
                ssl_options=dict(
                    certfile="foo.crt",
                    keyfile="foo.key",
                    cert_reqs=ssl.CERT_REQUIRED,
                    ca_certs="cacert.crt"))

        By default, the return value is a dictionary (or None, if no
        client certificate is present).  If ``binary_form`` is true, a
        DER-encoded form of the certificate is returned instead.  See
        SSLSocket.getpeercert() in the standard library for more
        details.
        http://docs.python.org/library/ssl.html#sslsocket-objects
        """
        try:
            return self.connection.stream.socket.getpeercert(
                binary_form=binary_form)
        except ssl.SSLError:
            return None



"""
This module implements a flexible global configuration system.


.. autoclass:: GConfig
    :members: reset_all_parameters, reset_parameter, write_parameters,
     write_few_parameters, read_parameters, read_parameter

.. autoexception:: GConfigTemplateError

.. autoexception:: GConfigValidateError

"""

from ginsfsm.globals import get_global_app
from ginsfsm.compat import (
    iteritems_,
    string_types,
    integer_types,
    binary_type,
    bytes_,
)

accepted_types = (str, int, bool, list, tuple, dict, bytes)


class GConfigTemplateError(Exception):
    """ Raised when something is wrong in the :term:`gconfig-template`
        definition.
    """


class GConfigValidateError(Exception):
    """ Raised when something is a parameter is not validated.
    """


def add_gconfig(gconfig, new_gconfig):
    """ Add to gconfig a new_gconfig """
    if new_gconfig is None:
        return gconfig

    if gconfig is not None:
        if isinstance(gconfig, (list, tuple)):
            gconfig = list(gconfig)
            gconfig.append(new_gconfig)
        else:
            gconfig = [gconfig, new_gconfig]
    else:
        gconfig = [new_gconfig]
    return gconfig


class _Config(object):
    """ Collect config properties.
    """


class GConfig(object):
    """Global configuration system.

    :param template: a dictionary or a list of dictionary
        describing the template of configuration parameters,
        with key/value as::

            'parameter_name':
            [type, default_value, flag, validate_function, description]

    Each parameter is defined with a template, that it's a list of 4 elements:

    * type: must be a type of:
      ``str``, ``int``, ``bool``, ``list``, ``dict`` or ``tuple``.
    * default_value: default value.
    * flag: modify behaviour.
    * validate_function: ``None`` or a ``callable``.
      The ``callable`` must return ``True`` if validates the value,
      otherwise ``False``. If the callable return ``False``
      a :exc:`GConfigValidateError` will be raised.
    * description: String describing the parameter.

    If the template is not valid, a :exc:`GConfigTemplateError` exception
    is raised.
    """

    FLAG_DIRECT_ATTR = 0x01  # write directly in gobj (not in config attr).

    def __init__(self, templates, logger=None):
        self._gconfig_template = {}
        self.config = _Config()
        self.logger = logger

        if not isinstance(templates, (list, tuple)):
            templates = [templates]

        for template in templates:
            if not issubclass(template.__class__, dict):
                raise GConfigTemplateError(
                    "GConfig(%r) template in %r is not a dict" %
                    (template, self.__class__.__name__))
            for parameter_name, definition in iteritems_(template):
                if not isinstance(parameter_name, string_types) or \
                        len(parameter_name) == 0:
                    raise GConfigTemplateError(
                        "Parameter name %r in %r is not a string" %
                        (parameter_name, self.__class__.__name__))
                if not isinstance(definition, (list, tuple)):
                    raise GConfigTemplateError(
                        "Parameter definition %r in %r is not list or tuple" %
                        (definition, self.__class__.__name__))
                if len(definition) != 5:
                    raise GConfigTemplateError(
                        "Parameter definition %r in %r is"
                        "  not a list or tuple of 5 items" %
                        (definition, self.__class__.__name__))
                type_, default, flag, validate, desc = definition
                if validate is not None and not callable(validate):
                    raise GConfigTemplateError(
                        "Parameter definition %r in %r:"
                        "  %r is not a callable" % (
                            definition, self.__class__.__name__, validate))
                if desc is not None and not isinstance(desc, string_types):
                    raise GConfigTemplateError(
                        "Parameter definition %r in %r:"
                        "  %r is not a string" %
                        (definition, self.__class__.__name__, desc))
                if type_ is not None and not issubclass(type_, accepted_types):
                    raise GConfigTemplateError(
                        "Parameter definition %r in %r:"
                        "  %r is not a type in %r" %
                        (definition, self.__class__.__name__,
                        type_, accepted_types))

                # [type, default_value, flag, validate_function, desc]

                self._gconfig_template.update({parameter_name: definition})

        # create default parameter values.
        for parameter, definition in iteritems_(self._gconfig_template):
            value = definition[1]
            self._write_parameter(parameter, value)

    def reset_all_parameters(self):
        """ Reset all parameters to default values.
        """
        kw = {}
        for parameter, definition in iteritems_(self._gconfig_template):
            kw.update({parameter: definition[1]})
        self.write_parameters(**kw)

    def reset_parameter(self, parameter):
        """ Reset one parameter to his default value.
        """
        definition = self._gconfig_template.get(parameter, None)
        if definition is not None:
            value = definition[1]
        kw = {parameter: value}
        self.write_parameters(**kw)

    def write_few_parameters(self, parameter_list, **kw):
        """ Write a few parameters.

        :param parameters: write only the parameters in ``parameter_list``.

        :param kw: keyword arguments with parameter_name:value pairs.

        .. warning:: Only the parameters defined in the template are writted,
            the remaining are ignored.
        """
        for parameter, value in iteritems_(kw):
            if parameter not in parameter_list:
                continue
            try:
                self._write_parameter(parameter, value)
            except Exception as e:
                # In real time don't raise exceptions, only at setup
                if self.logger:
                    self.logger.exception(e)

    def write_parameters(self, **kw):
        """ Write parameters.

        :param kw: keyword arguments with parameter_name:value pairs.

        .. warning:: Only the parameters defined in the template are writted,
            the remaining are ignored.
        """
        for parameter, value in iteritems_(kw):
            try:
                self._write_parameter(parameter, value)
            except Exception as e:
                # In real time don't raise exceptions, only at setup
                if self.logger:
                    self.logger.exception(e)

    def filter_parameters(self, **settings):
        """ Filter the parameters in settings belongs to gobj.
        The gobj is searched by his named-gobj or his gclass name.
        The parameter name in settings, must be a dot-named,
        with the first item being the named-gobj o gclass name.
        """
        parameters = {}
        named_gobj = self.name
        class_name = self.__class__.__name__

        for key, value in iteritems_(settings):
            names = key.rsplit('.', 1)
            if len(names) > 1:
                # Can be:
                #       GClass.attribute
                #       named-gobj.attribute
                name = names[0]
                attribute = names[1]
                if named_gobj and named_gobj == name:
                    # search by named-gobj
                    parameters[attribute] = value
                elif class_name == name:
                    # search by gclass name
                    parameters[attribute] = value
                elif name == 'GObj':
                    parameters[attribute] = value
            else:
                parameters[key] = value
        return parameters

    def _write_parameter(self, parameter, value):
        """ Write parameter. Raise and log the errors.
        Use: in setup we raise errors, in real time we log the errors.
        """
        # [type, default_value, flag, validate_function, desc]
        definition = self._gconfig_template.get(parameter, None)
        if definition is None:
            self.logger and self.logger.error(
                "ERROR %r in write_parameter(%r):"
                "\n  name = %r\n  value = %r" % (
                "PARAMETER NAME INVALID",
                self, parameter, value)
            )
            return

        validate = definition[3]
        type_ = definition[0]
        flag = definition[2]

        try:
            value = self._check(type_, value, validate)
        except Exception as e:
            msg = "ERROR %s in write_parameter(%r):" \
                "\n  name = %r\n  value = %r" % (
                e, self, parameter, value)
            self.logger and self.logger.exception(msg)
            raise GConfigValidateError(msg)

        if isinstance(value, string_types):
            # TODO: doc this!
            # TODO: move this to validate function
            prefix = value.split(':', 1)
            if prefix[0] == 'app':
                value = get_global_app(prefix[1])
                if not value:
                    self.logger and self.logger.error(
                        "ERROR get_global_app (%s) NOT LOADED" %
                        prefix[1])

        if flag and GConfig.FLAG_DIRECT_ATTR:
            destobj = self
        else:
            destobj = self.config

        if hasattr(destobj, parameter):
            attr = getattr(destobj, parameter)
            if callable(attr):
                if not callable(value):
                    # Override methods only with callables
                    if value is not None:
                        self.logger and self.logger.error(
                            "ERROR cannot override parameter (%r,%r)" % (
                                parameter, value))
                    return
        # OLD setattr(self, parameter, value)
        setattr(destobj, parameter, value)

    def _check(self, type_, value, validate):
        """ Check the value and convert into his definition type.
            With validate you can make special types.
        """
        if validate is not None:
            value = validate(value)

        if type_ is None:
            pass
        elif issubclass(type_, string_types):
            if value is not None:
                value = str(value)
        elif issubclass(type_, binary_type):
            if value is not None:
                value = bytes_(value)
        elif issubclass(type_, bool):  # first bool: it's a int type too!!
            value = asbool(value)
        elif issubclass(type_, integer_types):
            value = int(value)
        elif issubclass(type_, list):
            if isinstance(value, string_types):
                # this can have colateral effect
                value = value.replace(',', ' ')
                value = value.split()
            elif not hasattr(value, '__iter__'):
                value = [value, ]
            value = list(value)
        elif issubclass(type_, dict):
            value = dict(value)
        elif issubclass(type_, tuple):
            if isinstance(value, string_types):
                # this can have colateral effect
                value = value.replace(',', ' ')
                value = value.split()
            elif not hasattr(value, '__iter__'):
                value = (value,)
            value = tuple(value)
        else:
            raise ValueError('Unknown type %r' % value)
        return value

    def read_parameters(self):
        """ Return a dictionary with the current parameter:value pairs.
        """
        # OLD params = self.__dict__.copy()
        params = self.config.__dict__.copy()
        # OLD params.pop('_gconfig_template')
        return params

    def read_parameter(self, parameter):
        """ Return the current value of parameter.
        """
        # [type, default_value, flag, validate_function, desc]
        definition = self._gconfig_template.get(parameter, None)
        if not definition:
            # In real time, we log the errors instead of raise.
            self.logger and self.logger.error(
                "ERROR in %r, doesn't exist parameter (%r)" % (
                self, parameter))
            return None
        flag = definition[2]
        if flag and GConfig.FLAG_DIRECT_ATTR:
            read_in = self
        else:
            read_in = self.config

        try:
            # OLD value = getattr(self, parameter)
            value = getattr(read_in, parameter)
        except AttributeError:
            # In real time, we log the errors instead of raise.
            self.logger and self.logger.error(
                "ERROR in %r, doesn't exist parameter (%r)" % (
                self, parameter))
            return None
        return value


truthy = frozenset(('t', 'true', 'y', 'yes', 'on', '1'))


def asbool(s):
    """ Return the boolean value ``True`` if the case-lowered value of string
    input ``s`` is any of ``t``, ``true``, ``y``, ``on``, or ``1``, otherwise
    return the boolean value ``False``.  If ``s`` is the value ``None``,
    return ``False``.  If ``s`` is already one of the boolean values ``True``
    or ``False``, return it."""
    if s is None:
        return False
    if isinstance(s, bool):
        return s
    s = str(s).strip()
    return s.lower() in truthy



# -*- encoding: utf-8 -*-
""" GRouter GObj

.. autoclass:: GRouter
    :members:

"""

"""

GinsFSM forma parte de GObj-Ecosistema, un sistema donde el elemento basico
es un GObj, un objecto cuyo comportamiento esta definido por un automata,
y donde toda la comunicación con el objecto se realiza a traves de eventos.
El GObj define qué eventos de entrada soporta, y que eventos de salida ofrece.
A un GObj puedes enviarle un evento, o subscribirte a sus eventos de salida.

Puedes enviar y subscribir eventos de un GObj que esté en tu mismo thread,
en otros threads o subprocesos de tu aplicación, o en procesos corriendo
en la misma o diferente maquina. Esto incluye también a los browsers.

El transporte de los eventos se realiza mediante el protocolo websocket.
Cada thread o subproceso, representado por el object GAplic,
puede levantar un servicio websocket como
servidor o establecer tantas conexiones cliente como necesite (GRouter)
Se puede configurar un servidor websocket en cualquier puerto.
tambien en el puerto 80 o 443,
en este caso retringido a usar un servidor sockjs integrado
en el framework Pyramid, soportando ademas conexiones de browsers.

Para los GObj corriendo en browsers se usa la libreria sockjs. Esto permite
usar tanto navegadores modernos que soporten websocket, como navegadores
antiguos que no dispongan de websocket.

En el GObj-Ecosistema los threads o suprocesos estan representados por el
objecto GAplic, que además de ser el contenedor de todos los GObj del thread,
y suministrar el event loop, se puede definir con un nombre y uno varios roles.

Puedes configurar una red de gaplics, e intercambiar eventos entre ellos
simplemente conociendo el nombre o role del gaplic, y el nombre del gobj


NEW IDEAS:
Los eventos se garantiza que se entregan al siguiente nodo.
Si se ha podido entregar entonces llega el ack
que provoca que se borre de la cola.
Motivos para que no llege:
    destino desconectado,
    destinatario desconocido en el destino
    cola destinatario llena.

Las tarjetas de presentación son obligatorias si se quiere tener
eventos en sentido contrario, de retorno,
por ejemplo cuando se quiere recibir el return de la acción.

Las tarjetas de presentación no necesitan ack.
En cada conexión automáticamente se enviará la tarjeta de presentación
(solo los clientes: el cliente enviará al servidor conectado su tj.pr.)

Se podría tener un ack de la tr.pr.
que serviría para actualizar, comprobar, ampliar el nombre/roles
que disponemos de ese cliente.
El nombre y roles con que hemos configurado la url de una ruta estática,
no tiene porque coincidir con la que realmente tenga
el dueño de la ruta estática.

Para el sistema de alarmas es importante conocer los nak producidos,
donde se producen y el motivo.
Con esta información se podrán subsanar errores de configuración en el sistema.

Para el sistema de alarmas es importante conocer las desconexiones de las
rutas estáticas.
(CONECTADO/DESCONECTADO)
Con esta información se podrá poner en marcha un sistema de alerta
y comunicación a los afectados, para subsanar lo antes posible el problema.
En este caso, ya no es competencia exclusiva nuestra, sino que pueden
necesitar intervenir agentes externos.

Otro nivel superior de alarmas se puede establecer para informar
de la falta de acuses de recibo a nivel de mensajes de aplicación.
(PROVEEDOR NO CONTESTA)
No sé si se podrá establecer algo genérico para todos.

A nivel de estadísticas, eventos y bytes recibidos y transmitidos por
cada ruta (aparte de los estados de conexión, log activado, etc).

Los eventos podrán tener la propiedad de garantizados:
garantizados por un tiempo: finito o infinito.

Los eventos garantizados se enviarán al role Contenedor de garantizados.
En el Contenedor de garantizados, el estado optimo es todo vacio.
Las alarmas se fijarán en los niveles de llenado del contenedor.
El contenedor se encargará de borrar los eventos caducados e informar al
sistema de alarmas.

El contenedor de garantizados, podrá consultar al registro global de gobjs,
para ver si puede entregar los eventos.

El registro global de gobjs... todos los gaplics deberian registrarse alli.
"""

import datetime
from collections import deque
from ginsfsm.gobj import (
    GObj,
    GObjError,
)
from ginsfsm.c_srv_sock import GServerSock
from ginsfsm.c_connex import GConnex
from ginsfsm.c_timer import GTimer
from ginsfsm.protocols.sockjs.server.c_websocket import GWebSocket
from ginsfsm.protocols.sockjs.server.conn import SockJSConnection
from ginsfsm.protocols.sockjs.server.c_sockjs_server import GSockjsServer
from ginsfsm.utils import (
    get_host_port_from_urls,
    get_path_from_url,
    hexdump,
    random_key,
)
from ginsfsm.protocols.sockjs.server.proto import json_decode


class ExternalSubscription(object):
    """ Container of external subscription.
    """
    def __init__(self, intra_event, route_ref):
        self.event_name = intra_event.event_name
        self.destination_gaplic = intra_event.destination_gaplic
        self.destination_role = intra_event.destination_role
        self.destination_gobj = intra_event.destination_gobj
        self.origin_gaplic = intra_event.origin_gaplic
        self.origin_role = intra_event.origin_role
        self.origin_gobj = intra_event.origin_gobj
        self.route_ref = route_ref
        self.isubscription = None  # reference to internal subscription


class IntraEvent(object):
    """ Container of intra event.
    """
    event_fields = [
        'message_type',
        'serial',
        'event_name',
        'destination_gaplic',
        'destination_role',
        'destination_gobj',
        'origin_gaplic',
        'origin_role',
        'origin_gobj',
        'kw',
    ]
    event_ack_fields = [
        'message_type',
        'serial',
        'event_name',
        'reference',
    ]
    event_nack_fields = [
        'message_type',
        'serial',
        'event_name',
        'error_message',
    ]
    identity_card_fields = [
        'message_type',
        'my_gaplic_name',
        'my_roles',
    ]
    identity_card_ack_fields = [
        'message_type',
        'my_gaplic_name',
        'my_roles',
    ]

    def __init__(
            self,
            message_type=None,
            route_ref=None,
            serial=None,
            event_name=None,
            destination_gaplic=None,
            destination_role=None,
            destination_gobj=None,
            origin_gaplic=None,
            origin_role=None,
            origin_gobj=None,
            my_gaplic_name=None,
            my_roles=None,
            kw=None,
            error_message=None,
            reference=None):
        self.message_type = message_type
        self.route_ref = route_ref
        self.serial = serial
        self.event_name = event_name
        self.destination_gaplic = destination_gaplic
        self.destination_role = destination_role
        self.destination_gobj = destination_gobj
        self.origin_gaplic = origin_gaplic
        self.origin_role = origin_role
        self.origin_gobj = origin_gobj
        self.my_gaplic_name = my_gaplic_name
        self.my_roles = my_roles
        if kw:
            self.kw = kw
        else:
            self.kw = {}
        self.error_message = error_message
        self.reference = reference

    def toJSON(self):
        message_type = self.message_type
        ret = {}
        fn = lambda f: ret.update({f: getattr(self, f)})
        if message_type == '__event__':
            list(map(fn, self.event_fields))
        elif message_type == '__event_ack__':
            list(map(fn, self.event_ack_fields))
        elif message_type == '__event_nack__':
            list(map(fn, self.event_nack_fields))
        elif message_type == '__identity_card__':
            list(map(fn, self.identity_card_fields))
        elif message_type == '__identity_card_ack__':
            list(map(fn, self.identity_card_ack_fields))
        else:
            self.logger.info(
                'ERROR unknown message type %r' % message_type)
        return ret

    def copy(self, new_message_type):
        """ Copy the intra event to new message type
            with only the necessary fields to new type.
        """
        fields = {}
        fn = lambda f: fields.update({f: getattr(self, f)})
        if new_message_type == '__event__':
            list(map(fn, self.event_fields))
        elif new_message_type == '__event_ack__':
            list(map(fn, self.event_ack_fields))
        elif new_message_type == '__event_nack__':
            list(map(fn, self.event_nack_fields))
        elif new_message_type == '__identity_card__':
            list(map(fn, self.identity_card_fields))
        elif new_message_type == '__identity_card_ack__':
            list(map(fn, self.identity_card_ack_fields))
        else:
            self.logger.info(
                'ERROR unknown message type %r' % new_message_type)
        fields.update({'message_type': new_message_type})
        new = IntraEvent(**fields)
        return new

    def __repr__(self):
        return "%r" % self.toJSON()


class StaticRoute(object):
    """ Static route (when we are clients)
    """
    def __init__(self, name, roles, urls, connex):
        self.gaplic_name = name
        if isinstance(roles, (list, tuple,)):
            self.roles = roles
        elif roles:
            self.roles = [roles]
        else:
            self.roles = []
        if isinstance(urls, (list, tuple)):
            self.urls = urls
        else:
            self.urls = [urls]
        if not self.urls:
            raise ValueError(
                'StaticRoute needs urls'
            )

        connex.route = self
        self.connex = connex
        self.idx_url = 0        # current url connection if client
        self.dl_output_events = deque()  # queue of pending output events
        self.cur_pending_event = None
        self.write = None
        self.route_ref = hash(connex)
        self.identity_ack = False
        self.serial = 0
        self.ws = None
        self.their_roles = None  # their roles anwered in identity ack (with *)

    def incr_serial(self):
        self.serial += 1
        if self.serial > 10:
            self.serial = 1
        return self.serial

    def __str__(self):
        return "%s %s" % (self.gaplic_name, self.route_ref)

    def __repr__(self):
        return "\nSTATIC ROUTE:\n  gaplic_name: %s\n  roles: %s\n  " \
            "route_ref: %s\n  connex: %s\n  their_roles: %s\n" % (
                self.gaplic_name,
                self.roles,
                self.route_ref,
                self.connex.config.destinations,
                self.their_roles,
            )


class DynamicRoute(object):
    """ Dynamic route (when we are servers)
    """
    def __init__(self, name, roles, write, gsock):
        self.gaplic_name = name
        self.roles = roles
        if isinstance(roles, (list, tuple,)):
            self.roles = roles
        elif roles:
            self.roles = [roles]
        else:
            self.roles = []
        self.dl_output_events = deque()  # queue of pending output events
        self.cur_pending_event = None
        self.connex = None
        self.write = write
        self.gsock = gsock
        self.route_ref = hash(gsock)
        self.identity_ack = True
        self.serial = 0
        self.ws = None
        self.their_roles = None  # their roles anwered in identity (with ^)

    def incr_serial(self):
        self.serial += 1
        if self.serial > 10:
            self.serial = 1
        return self.serial

    def __str__(self):
        return "%s %s" % (self.gaplic_name, self.route_ref)

    def __repr__(self):
        return "\DINAMYC ROUTE:\n  gaplic_name: %s\n  roles: %s\n  " \
            "route_ref: %s\n  gsock: %r\n  their_roles: %r\n" % (
                self.gaplic_name,
                self.roles,
                self.route_ref,
                self.gsock,
                self.their_roles,
            )


class EmptyRoute(object):
    """ Empty route
    """
    def __init__(self):
        self.route_ref = None
        self.identity_ack = False

    def __str__(self):
        return "%s" % (self.route_ref)

    def __repr__(self):
        return "\EMPTY ROUTE:\n  route_ref: %s\n" % (
            self.route_ref,
        )


class GAplicRegistry(object):
    """ GAplics registry.
    """
    def __init__(self, router):
        self.router = router
        self.my_gaplic_name = router.gaplic.name  # self gaplic name
        self.my_roles = router.gaplic.roles  # self role
        self.static_routes = {}
        self.dynamic_routes = {}
        self.broadcasted_roles = []

    def add_static_route(self, name, roles, urls, connex_mode=None):
        if not isinstance(urls, (list, tuple)):
            urls = [urls]
        destinations = get_host_port_from_urls(urls)
        if not destinations:
            return None
        connex = self.router.create_gobj(
            'static_route',
            GConnex,
            self.router,
            transmit_ready_event_name=None,
            destinations=destinations,
        )
        r = StaticRoute(name, roles, urls, connex)
        self.static_routes[r.route_ref] = r
        return r

    def add_dynamic_route(self, name, role, write, gsock):
        if not name:
            # anonymous gaplic name
            name = "^%s%s" % (gsock.get_peername(), gsock.get_sockname())
        r = DynamicRoute(name, role, write, gsock)
        self.dynamic_routes[r.route_ref] = r
        return r

    def get_route_by_ref(self, route_ref):
        route = self.static_routes.get(route_ref, None)
        if not route:
            route = self.dynamic_routes.get(route_ref, None)
        return route

    def search_route(self, gaplic_name, role):
        registry = self

        #------------------------------#
        #   Search by gaplic_name
        #------------------------------#

        # search in static routes
        static_routes = registry.static_routes
        for route_ref in static_routes:
            route = static_routes[route_ref]
            if route.gaplic_name == gaplic_name:
                return route

        # search in dynamic routes
        dynamic_routes = registry.dynamic_routes
        for route_ref in dynamic_routes:
            route = dynamic_routes[route_ref]
            if route.gsock.is_disconnected():
                continue
            if route.gaplic_name == gaplic_name:
                return route

        #------------------------------#
        #   Search by role
        #------------------------------#

        # search in static routes
        static_routes = registry.static_routes
        for route_ref in static_routes:
            route = static_routes[route_ref]
            roles = route.roles
            for ro in roles:
                if ro == role:
                    return route

        # search in dynamic routes
        dynamic_routes = registry.dynamic_routes
        for route_ref in dynamic_routes:
            route = dynamic_routes[route_ref]
            if route.gsock.is_disconnected():
                continue
            roles = route.roles
            # TODO: busca el rol conectado si hay varios
            for ro in roles:
                ro = ro.lstrip('^')  # TODO: search removing ^ one to one
                if ro == role:
                    return route

        return None

    def enqueue_pending_event(self, route, intra_event):
        dl_output_events = route.dl_output_events
        ln = len(dl_output_events)
        if ln > self.router.config.limit_pending_output:
            self.router.logger.error("ERROR output event queue FULL: %d" % ln)
            self.check_route(route)
            return False
        dl_output_events.append(intra_event)
        return True

    def check_route(self, route):
        if isinstance(route, DynamicRoute):
            if route.gsock.is_closed() or route.gsock.is_disconnected():
                self.delete_route(route.route_ref)

    def delete_route(self, route_ref):
        key = route_ref
        dicc = self.dynamic_routes
        if key in dicc:
            self.router.logger.info("DELETING dynamic ROUTER %r" % (key))
            route = dicc[key]
            if route.ws:
                GObj.destroy_gobj(route.ws)
                route.ws = None
            del dicc[key]
            return True
        dicc = self.static_routes
        if key in dicc:
            self.router.logger.info("DELETING static ROUTER %r" % (key))
            route = dicc[key]
            if route.ws:
                GObj.destroy_gobj(route.ws)
                route.ws = None
            del dicc[key]
            return True
        return False

    def fire_pending_events(self, route):
        if not route.route_ref:
            return
        write = route.write
        trace_router = self.router.config.trace_router

        if not write:
            if trace_router:
                self.router.logger.info("    * route DISCONNECTED!!")
            return False

        if not route.identity_ack:
            # TODO: improve sending identity card with timeouts
            self.router.send_identity_card(route)
            return

        if route.cur_pending_event:
            try:
                msg = route.cur_pending_event.toJSON()
            except Exception as e:
                self.router.logger.exception(e)
                route.cur_pending_event = None  # trash message
                return False
            if trace_router:
                prefix = '%s ==> %s' % (
                    self.my_gaplic_name, route)
                self.router.trace_intra_event(prefix, route.cur_pending_event)
            write(msg)
            return True

        dl_output_events = route.dl_output_events
        if len(dl_output_events) > 0:
            route.cur_pending_event = dl_output_events.popleft()
            try:
                msg = route.cur_pending_event.toJSON()
            except Exception as e:
                self.router.logger.exception(e)
                route.cur_pending_event = None  # trash message
                return False
            if trace_router:
                prefix = '%s ==> %s' % (
                    self.my_gaplic_name, route)
                self.router.trace_intra_event(prefix, route.cur_pending_event)
            write(msg)
            return True
        return False

    def subcontracted_roles(self):
        """ Roles of static routes are our subcontracted roles
        """
        registry = self
        found_roles = set()

        # search in static routes
        static_routes = registry.static_routes
        for route_ref in static_routes:
            route = static_routes[route_ref]
            for ro in route.roles:
                found_roles.add(ro)

        found_roles = list(found_roles)
        return tuple(map(lambda x: '*' + x, found_roles))

    def get_roles_back_way(self):
        registry = self
        found_roles = set()

        # search in dynamic routes
        dynamic_routes = registry.dynamic_routes
        for route_ref in dynamic_routes:
            route = dynamic_routes[route_ref]
            if route.gsock.is_disconnected():
                continue
            roles = route.roles
            for ro in roles:
                found_roles.add(ro)
        found_roles = list(found_roles)
        return tuple(map(lambda x: '^' + x, found_roles))

    def check_broadcast_roles_back_way(self, roles):
        #broadcasted_roles = self.broadcasted_roles  # TODO: must be per route
        broadcasted_roles = []
        to_broadcast = []

        pure_roles = list(
            map(lambda x: x.lstrip('^'), roles)
        )

        for idx, ro in enumerate(pure_roles):
            if ro in self.my_roles:
                continue  # avoid loops
            if not ro in broadcasted_roles:
                to_broadcast.append('^' + roles[idx])
                broadcasted_roles.append(ro)

        if to_broadcast:
            self.broadcast_role_back_way(to_broadcast)

    def broadcast_role_back_way(self, roles_back_way):
        registry = self
        if self.router.config.trace_router:
            self.router.logger.info(
                "BROADCAST BACK WAY ====> %r" % roles_back_way)
        # search in static routes
        static_routes = registry.static_routes
        for route_ref in static_routes:
            route = static_routes[route_ref]
            if route.connex.is_disconnected():
                continue
            self.router.send_identity_card(route, roles_back_way)

    def role_level(self, role):
        role = role.strip()
        level = 0
        for level, char in enumerate(role):
            if not role[level] in ('^', '*'):
                break
        return level, role[level:]


class RouterConnection(SockJSConnection):
    """ Websocket application to work with Pyramid router.

    The messages come by here when the router server is a Pyramid web server.
    """
    clients = set()

    def on_open(self, info):
        session = self.session
        if session:
            session.write = session.send_message
            session.gsock = session.handler.gsock

    def on_message(self, msg):
        session = self.session
        router = session.gaplic.router
        try:
            msg = json_decode(msg)
        except Exception as e:
            router.logger.exception("ERROR %s:\n%s" % (e, hexdump('<==', msg)))
            return
        router.send_event(
            router,
            'EV_INPUT_MESSAGE',
            msg=msg,
            session=session,
        )

    def on_close(self):
        session = self.session
        router = session.gaplic.router
        route = session.route
        if route:
            router.registry.delete_route(route.route_ref)
        else:
            pass  # TODO: delete gsock or channel


def get_first_name(name):
    if '^' in name:
        name = name.split()[0]
    return name


def ac_connected(self, event):
    """ Someone is connected.
    By here we receive the connection of both client and server routers
    that are not pyramid routes.
    The GWebSocket gobj will take possession of all their events.
    """
    gsock = event.gsock
    source_gobj = event.source[-1]
    if hasattr(event, 'connex'):
        # client connection
        connex = event.connex
        route_ref = hash(connex)
        route = self.registry.get_route_by_ref(route_ref)
        # all urls must have the same path!!
        resource = get_path_from_url(route.urls[0])
        if not resource:
            resource = '/'
        ws = self.create_gobj(
            'websock_cli-' + source_gobj.name,
            GWebSocket,
            self,
            gsock=connex,
            iam_server=False,
            subscriber=self,
            resource=resource,
        )
        ws.route_ref = route_ref
        ws.route = route
        route.ws = ws
        route.write = ws.send_jsonfied_message

    else:
        # clisrv connection
        ws = self.create_gobj(
            'websock_srv-' + source_gobj.name,
            GWebSocket,
            self,
            gsock=gsock,
            iam_server=True,
            subscriber=self,
        )
        ws.route_ref = None
        ws.route = None


def ac_disconnected(self, event):
    """
    """


def ac_add_static_route(self, event):
    """ Add external gaplic.
    """
    name = event.kw.get('name', None)
    roles = event.kw.get('roles', '')
    urls = event.urls
    connex_mode = event.kw.get('connex_mode', None)

    route = self.registry.add_static_route(name, roles, urls, connex_mode)
    if not route:
        return -1
    if self.config.trace_router:
        self.logger.info("%s: NEW STATIC route: %r" % (
            self.registry.my_gaplic_name,
            route
        ))
    return route.route_ref


def ac_on_open(self, event):
    """
    By here we receive from both client and server routers
    that are not pyramid routes.
    """
    ws = event.source[-1]
    ws.write = ws.send_jsonfied_message

    # send our identity card if iam client
    # we only send the identity card in static routes!
    if not ws.config.iam_server:
        self.send_identity_card(ws.route)


def ac_on_close(self, event):
    """
    By here we receive from both client and server routers
    that are not pyramid routes.
    """
    ws = event.source[-1]
    route = ws.route
    if isinstance(route, DynamicRoute):
        self.registry.delete_route(route.route_ref)


def ac_on_message(self, event):
    """
    By here we receive from both client and server routers
    that are not pyramid routes.
    """
    ws = event.source[-1]
    msg = event.data

    try:
        msg = json_decode(msg)
    except Exception as e:
        self.logger.exception(
            "ERROR json_decode %s:\n%s" % (e, hexdump('<==', msg)),
        )
        return
    self.send_event(
        self,
        'EV_INPUT_MESSAGE',
        msg=msg,
        websocket=ws,
    )


def ac_input_message(self, event):
    """ Event attributes:
        * :attr:`msg`: received message.
    """
    registry = self.registry
    msg = event.msg
    try:
        intra_event = IntraEvent(**msg)
    except Exception as e:
        self.logger.error(
            "ERROR %s:\n%s" % (e, hexdump('<==', msg)),
            exc_info=True,
        )
        return
        raise
    message_type = intra_event.message_type
    trace_router = self.config.trace_router
    this_route = EmptyRoute()
    session = event.kw.get('session', None)  # entry by pyramid
    websocket = event.kw.get('websocket', None)  # entry by direct websocket
    if session:
        write = session.write
        gsock = session.gsock
        if hasattr(session, 'route'):
            if session.route:
                this_route = session.route
    elif websocket:
        write = websocket.write
        gsock = websocket.gsock
        if websocket.route:
            this_route = websocket.route

    if trace_router:
        prefix = '%s <== %s' % (
            registry.my_gaplic_name,
            this_route,
        )
        self.trace_intra_event(prefix, intra_event)

    if message_type == '__identity_card__':
        # Register a dynamic router
        gaplic_name = intra_event.my_gaplic_name
        roles = intra_event.my_roles
        if isinstance(this_route, EmptyRoute):
            # new dynamic route
            case = 'NEW'
            this_route = registry.add_dynamic_route(
                gaplic_name,
                roles,
                write,
                gsock,
            )
            if session:
                session.route = this_route
            elif websocket:
                websocket.route = this_route
                this_route.ws = websocket
            else:
                self.logger.error("ERROR no session neither websocket")
        else:
            # update dynamic route
            case = 'UPDATE'
            this_route.roles = roles  # TODO: check roles?

        if self.config.trace_router:
            self.logger.info("%s: %s DYNAMIC route: %s: %s: %s: %s" % (
                registry.my_gaplic_name,
                case,
                this_route.route_ref,
                this_route.gaplic_name,
                this_route.roles,
                gsock.addr))

        ack = IntraEvent(
            message_type='__identity_card_ack__',
            my_gaplic_name=registry.my_gaplic_name,
            my_roles=registry.my_roles + registry.subcontracted_roles(),
        )
        if trace_router:
            prefix = '%s ==> %s' % (
                registry.my_gaplic_name, this_route)
            self.trace_intra_event(prefix, ack)
        write(ack.toJSON())

        #
        #   Broadcast forward the new role's back ways.
        #
        registry.check_broadcast_roles_back_way(this_route.roles)

        #
        #   Somebody can be interested in know new clients
        #
        if case == 'NEW':
            self.broadcast_event(
                'EV_NEW_DYNAMIC_ROUTE',
                gaplic_name=gaplic_name,
                roles=roles,
            )

        return

    elif message_type == '__identity_card_ack__':
        # TODO: check if the info we have is correct
        this_route.identity_ack = True
        if intra_event.my_gaplic_name:
            if this_route.gaplic_name != intra_event.my_gaplic_name:
                self.logger.info(
                    "WARNING: updating route %r, GAPLIC_NAME! to %r" % (
                    this_route.route_ref,
                    intra_event.my_gaplic_name))
                this_route.gaplic_name = intra_event.my_gaplic_name

        # check their roles
        for r in intra_event.my_roles:
            if not r in this_route.roles:
                if self.config.update_unknown_roles:
                    self.logger.warn(
                        "WARNING: updating route %r, adding ROLE %r" % (
                        this_route.route_ref,
                        r))
                    this_route.roles.append(r)

        # check if our subcontracted roles are all
        for r in this_route.roles:
            pure_roles = list(
                map(lambda x: x.lstrip('*'), intra_event.my_roles)
            )
            if not r in pure_roles:
                self.logger.error(
                    "ERROR: %r LACKS role %r, it has only %r" % (
                    this_route,
                    r,
                    intra_event.my_roles))
        this_route.their_roles = intra_event.my_roles[:]

        if self.config.trace_router:
            self.logger.info("%s: UPDATE STATIC route: %r" % (
                self.registry.my_gaplic_name,
                this_route
            ))

        registry.fire_pending_events(this_route)
        return

    elif message_type == '__event_ack__' or message_type == '__event_nack__':
        if this_route.route_ref:
            cur_pending_event = this_route.cur_pending_event
            if not cur_pending_event:
                self.logger.error("ERROR processing ack: NO cur_pending_event")

            if cur_pending_event and \
                    intra_event.serial != cur_pending_event.serial:
                self.logger.error("ERROR processing ack: DIFFERENT SERIAL")

            del this_route.cur_pending_event
            this_route.cur_pending_event = None
            registry.fire_pending_events(this_route)
        return

    elif message_type == '__event__':
        #--------------------------------------------------#
        #   It's an event. Search if for us or to resend
        #--------------------------------------------------#
        own = False

        if isinstance(this_route, DynamicRoute):
            if not intra_event.origin_gaplic:
                #  Set our extended gaplic name
                intra_event.origin_gaplic = this_route.gaplic_name

        # search by gaplic
        if intra_event.destination_gaplic and registry.my_gaplic_name:
            first_gaplic_name = get_first_name(intra_event.destination_gaplic)
            if first_gaplic_name == registry.my_gaplic_name:
                own = True

        # search by role
        if intra_event.destination_role and registry.my_roles:
            if intra_event.destination_role in registry.my_roles:
                own = True

        if own:
            #------------------------------------#
            #   The event is for own gaplic.
            #------------------------------------#
            dst_gobj = intra_event.destination_gobj
            try:
                named_gobj = self.gaplic.find_unique_gobj(dst_gobj)
            except GObjError as e:
                ack = intra_event.copy('__event_nack__')
                ack.error_message = e
                if trace_router:
                    prefix = '%s ==> %s' % (
                        registry.my_gaplic_name, this_route)
                    self.trace_intra_event(prefix, ack)
                write(ack.toJSON())
                return

            if not named_gobj:
                ack = intra_event.copy('__event_nack__')
                ack.error_message = "GObj %r UNKNOWN" % dst_gobj
                if trace_router:
                    prefix = '%s ==> %s' % (
                        registry.my_gaplic_name, this_route)
                    self.trace_intra_event(prefix, ack)
                write(ack.toJSON())
                return

            if trace_router:
                self.logger.info("   !!! IntraEvent MINE !!!")

            if intra_event.kw.get('__subscribe_event__', False):
                # it's a external subscription
                external_subscriptor_ref = self.make_external_subscription(
                    intra_event,
                    this_route.route_ref
                )
                if not external_subscriptor_ref:
                    ack = intra_event.copy('__event_nack__')
                    ack.error_message = "Cannot make external subscription"
                    if trace_router:
                        prefix = '%s ==> %s' % (
                            registry.my_gaplic_name, this_route)
                        self.trace_intra_event(prefix, ack)
                    write(ack.toJSON())
                    return

                subs_kw = {
                    '__rename_event_name__': 'EV_SUBSCRIPTION',
                    '__subscription_reference__': external_subscriptor_ref,
                }
                try:
                    isubscription = named_gobj.subscribe_event(
                        intra_event.event_name,
                        self,
                        **subs_kw
                    )
                except Exception as e:
                    self.remove_external_subscription(external_subscriptor_ref)
                    ack = intra_event.copy('__event_nack__')
                    ack.error_message = "Cannot subscribe event: %s" % e
                    if trace_router:
                        prefix = '%s ==> %s' % (
                            registry.my_gaplic_name, this_route)
                        self.trace_intra_event(prefix, ack)
                    write(ack.toJSON())
                    return
                esubscription = self.get_external_subscription(
                    external_subscriptor_ref)
                esubscription.isubscription = isubscription

                # simple ack, no checkout, for the sender could remove the msg.
                ack = intra_event.copy('__event_ack__')
                ack.reference = external_subscriptor_ref
                if trace_router:
                    prefix = '%s ==> %s' % (
                        registry.my_gaplic_name, this_route)
                    self.trace_intra_event(prefix, ack)
                write(ack.toJSON())
                return

            if intra_event.kw.get('__unsubscribe_event__', False):
                # it's a external unsubscription
                subscription_reference = intra_event.kw.get(
                    '__subscription_reference__', None)

                if subscription_reference:
                    self.remove_external_subscription(subscription_reference)
                else:
                    # TODO: find and delete external subscription
                    self.logger.error("ERROR no __subscription_reference__")

                # simple ack, no checkout, for the sender could remove the msg.
                ack = intra_event.copy('__event_ack__')
                if trace_router:
                    prefix = '%s ==> %s' % (
                        registry.my_gaplic_name, this_route)
                    self.trace_intra_event(prefix, ack)
                write(ack.toJSON())
                return

            # simple ack, no checkout, for the sender could remove the msg.
            ack = intra_event.copy('__event_ack__')
            if trace_router:
                prefix = '%s ==> %s' % (
                    registry.my_gaplic_name, this_route)
                self.trace_intra_event(prefix, ack)
            write(ack.toJSON())

            # inject the event in the gaplic
            ret = self.send_event(
                named_gobj,
                intra_event.event_name,
                __intra_event__=intra_event,
                **intra_event.kw
            )

            # send action resp as event if they subscribe it.
            if intra_event.kw.get('__subscribe_response__', False):
                resp = IntraEvent(
                    message_type='__event__',
                    route_ref=None,
                    serial=this_route.incr_serial(),
                    event_name=intra_event.event_name,
                    destination_gaplic=intra_event.origin_gaplic,
                    destination_gobj=intra_event.origin_gobj,
                    origin_gaplic=registry.my_gaplic_name,
                    origin_gobj=named_gobj.name,
                    kw={
                        '__action_return__': ret,
                    }
                )

                registry.enqueue_pending_event(
                    this_route,
                    resp,
                )
                if not this_route.cur_pending_event:
                    registry.fire_pending_events(this_route)

            return

        #-------------------------------------------#
        #       The event is for resend.
        #       Search in routes.
        #-------------------------------------------#
        route_dst = registry.search_route(
            intra_event.destination_gaplic,
            intra_event.destination_role)

        if not route_dst:
            # nothing found
            ack = intra_event.copy('__event_nack__')
            ack.error_message = "Destination gaplic %r or role % r" \
                " NOT FOUND" % (
                    intra_event.destination_gaplic,
                    intra_event.destination_role
                )
            if trace_router:
                prefix = '%s ==> %s' % (
                    registry.my_gaplic_name, this_route)
                self.trace_intra_event(prefix, ack)
            write(ack.toJSON())
            return

        if trace_router:
            self.logger.info("    * IntraEvent FORWARD!!")

        if not route_dst.write:
            # nothing found
            ack = intra_event.copy('__event_nack__')
            ack.error_message = "Route found but DISCONNECTED"
            if trace_router:
                prefix = '%s ==> %s' % (
                    registry.my_gaplic_name, this_route)
                self.trace_intra_event(prefix, ack)
            write(ack.toJSON())
            return

        # enqueue the action response event
        ret = registry.enqueue_pending_event(
            route_dst,
            intra_event,
        )
        # simple ack, no checkout, for the sender could remove the msg.
        if ret:
            ack = intra_event.copy('__event_ack__')
        else:
            ack = intra_event.copy('__event_nack__')
            if trace_router:
                self.logger.info("    * event queue FULL!!")
            ack.error_message = "route_dst event queue FULL"
        if trace_router:
            prefix = '%s ==> %s' % (
                registry.my_gaplic_name, this_route)
            self.trace_intra_event(prefix, ack)
        write(ack.toJSON())

        if not route_dst.cur_pending_event:
            registry.fire_pending_events(route_dst)
        else:
            if ret:
                if trace_router:
                    self.logger.info("    * route_dst BUSY!!")

    else:
        self.logger.error("=================================================")
        self.logger.error("ERROR message_type UNKNOWN: %r" % message_type)


def ac_timeout(self, event):
    something_todo = self.check_routes()
    if something_todo:
        self.set_timeout(self.config.timeout_idle)


def ac_show_routes(self, event):
    ret = self.mt_show_routes()
    return ret


def ac_command(self, event):
    command = event.command
    ret = self.mt_execute_command(command)
    return ret


def ac_subscription(self, event):
    subscription_reference = event.__subscription_reference__
    subscription = self.external_subscriptions.get(
        subscription_reference, None)
    if subscription:
        self._send_event_to_external_route(
            subscription.route_ref,
            subscription.origin_gaplic,
            subscription.origin_role,
            subscription.origin_gobj,
            subscription.event_name,
            subscription.destination_gobj,
            event.kw)


def ac_send_event_outside(self, event):
    gaplic_name = event.gaplic_name
    role = event.role
    gobj_name = event.gobj_name
    event_name_ = event.event_name_
    subscriber_gobj = event.subscriber_gobj
    origin_role = event.origin_role
    kw = event.kw

    registry = self.registry
    route = registry.search_route(gaplic_name, role)
    if not route:
        self.logger.error(
            "ERROR ac_send_event_outside() "
            "route NOT FOUND: %r %r" % (gaplic_name, role))
        return False

    intra_event = IntraEvent(
        message_type='__event__',
        serial=route.incr_serial(),
        event_name=event_name_,
        destination_gaplic=gaplic_name,
        destination_role=role,
        destination_gobj=gobj_name,
        origin_gaplic=registry.my_gaplic_name,
        origin_role=origin_role if origin_role else registry.my_roles[0],
        origin_gobj=subscriber_gobj,
        kw=kw)

    registry.enqueue_pending_event(route, intra_event)
    if not route.cur_pending_event:
        registry.fire_pending_events(route)
    return True


GROUTER_FSM = {
    'event_list': (
        'EV_ADD_STATIC_ROUTE: top input',
        'EV_COMMAND: top input',
        'EV_SHOW_ROUTES: top input',
        'EV_CONNECTED: bottom input',
        'EV_DISCONNECTED: bottom input',
        'EV_ON_OPEN: bottom input',
        'EV_ON_CLOSE: bottom input',
        'EV_ON_MESSAGE: bottom input',
        'EV_INPUT_MESSAGE: bottom input',
        'EV_TIMEOUT: bottom input',
        'EV_SUBSCRIPTION: bottom input',
        'EV_SEND_EVENT_OUTSIDE: top input',
        'EV_NEW_DYNAMIC_ROUTE: top output',
    ),
    'state_list': ('ST_IDLE',),
    'machine': {
        'ST_IDLE':
        (
            ('EV_COMMAND',              ac_command,             None),
            ('EV_SHOW_ROUTES',          ac_show_routes,         None),
            ('EV_ADD_STATIC_ROUTE',     ac_add_static_route,    None),
            ('EV_CONNECTED',            ac_connected,           None),
            ('EV_DISCONNECTED',         ac_disconnected,        None),
            ('EV_ON_OPEN',              ac_on_open,             None),
            ('EV_ON_CLOSE',             ac_on_close,            None),
            ('EV_ON_MESSAGE',           ac_on_message,          None),
            ('EV_INPUT_MESSAGE',        ac_input_message,       None),
            ('EV_TIMEOUT',              ac_timeout,             None),
            ('EV_SUBSCRIPTION',         ac_subscription,        None),
            ('EV_SEND_EVENT_OUTSIDE',   ac_send_event_outside,  None),
        ),
    }
}


def validate_static_routes(value):
    """
    sca1, sca+sca, http://localhost:8002+http://localhost:8002;
    sca2, sca+sca, http://localhost:8003+http://localhost:8003;

    must convert in:

    ((
        'sca1',
        ('sca','sca'),
        ('http://localhost:8002', 'http://localhost:8002'),
    ),
    (
        'sca2',
        ('sca','sca'),
        ('http://localhost:8003', 'http://localhost:8003'),
    ))

    """
    routes = []
    if value:
        # save each route separated by ;
        value = value.split(';')
        for r in value:
            r = r.strip()
            if r:
                routes.append(r)

        # split each route by ,
        for idx, r in enumerate(routes):
            items = r.split(',')
            v = []
            for it in items:
                it = it.strip()
                if it:
                    v.append(it)
            routes[idx] = v

        for r in routes:
            for idx, field in enumerate(r):
                v = ()
                if not '+' in field:
                    v += (field,)
                    continue
                items = field.split('+')
                for e in items:
                    e = e.strip()
                    if e:
                        v += (e,)
                r[idx] = v

    return routes


GROUTER_GCONFIG = {
    #TODO: dns router with a short range ports?
    'server': [
        bool, False, 0, None,
        'Server router. Any gaplic with public services must be server'
    ],
    'localhost_route_ports': [
        tuple, range(2690, 2700), 0, None,
        'List of posible listening ports acting as localhost tcp router server'
    ],
    'pyramid_router_url': [
        str, '__pyramid_router__', 0, None,
        'pyramid url to start the router'
    ],
    'pyramid_root': [None, None, 0, None, 'pyramid root (parent)'],
    'trace_router': [bool, False, 0, None, "trace route messages"],
    'limit_pending_output': [int, 2, 0, None, "output queue size"],
    'timeout_idle': [int, 5, 0, None, "idle timeout"],
    'static_routes': [
        tuple, None, 0, validate_static_routes,
        'static routes with format:'
        'name, role1+role2, url1+url2; name2, role2, url2; ...'
    ],
    'update_unknown_roles': [bool, False, 0, None, "update unknown roles"],
}


class GRouter(GObj):
    """  GRouter GObj.
    Exchange events between GAplic's of same or different hosts.
    This gobj is permanent and direct child of his GAplic.

    The GRouter have two options:

    * open an own listening port,
    * use a url in a http server environment.

    The first option only works with a pure websocket channel.
    The second option needs an application done with Pyramid framework
    and implements sockjs-protocol.

    .. ginsfsm::
       :fsm: GROUTER_FSM
       :gconfig: GROUTER_GCONFIG

    *Input-Events:*
        * :attr:`'EV_TIMEOUT'`: Timer over.
            Start the machine.

        * :attr:`'EV_CONNECTED'`: New client.

        * :attr:`'EV_DISCONNECTED'`: Client disconnected.

    *Output-Events:*

    """
    def __init__(self):
        GObj.__init__(self, GROUTER_FSM, GROUTER_GCONFIG)
        self.server_sock = None  # server sock of tcp router
        self.local_router = None
        self.external_subscriptions = {}

    def start_up(self):
        """ Initialization zone.
        """
        self.registry = GAplicRegistry(self)
        if self.config.pyramid_router_url:
            # router inside pyramid.
            self.sockjs = self.create_gobj(
                self.config.pyramid_router_url,       # named gobj
                GSockjsServer,          # gclass
                self.config.pyramid_root,      # parent
                sockjs_app_class=RouterConnection,
                response_limit=4096,
                #disabled_transports=('websocket',)  # we need support IE8
            )
        elif self.config.server:
            # router with server sock
            self.start_router_server()

        self._timer = self.create_gobj(
            None,
            GTimer,
            self,
            timeout_event_name='EV_TIMEOUT')
        self.set_timeout(self.config.timeout_idle)

        if self.config.static_routes:
            routes = self.config.static_routes
            for r in routes:
                name, roles, urls = r
                self.send_event(
                    self,
                    'EV_ADD_STATIC_ROUTE',
                    name=name,
                    roles=roles,
                    urls=urls,
                )

    def set_timeout(self, seconds):
        if self._timer:  # protect from overwrite_parameters
            self.send_event(self._timer, 'EV_SET_TIMER', seconds=seconds)

    def clear_timeout(self):
        if self._timer:  # protect from overwrite_parameters
            self.send_event(self._timer, 'EV_SET_TIMER', seconds=-1)

    def mt_execute_command(self, command):
        ret = 'EXECUTED ' + command + ' Ok'
        return ret

    def mt_show_routes(self):
        registry = self.registry
        s = 'My gaplic name: %r\n' % registry.my_gaplic_name
        s += 'My roles: %r\n\n' % registry.my_roles
        static_routes = registry.static_routes
        s += 'Static routes:\n'
        for key in static_routes:
            r = static_routes[key]
            name = r.gaplic_name
            roles = r.roles
            urls = r.urls
            acked = 'Yes Acked' if r.identity_ack else 'NOT Acked'
            s += '  route_ref: %-12s | name:%r | roles:%r | %s | %s\n' % (
                key,
                name,
                roles,
                urls,
                acked,
            )
        s += '\n'

        s += 'Dynamic routes:\n'
        dynamic_routes = registry.dynamic_routes
        for key in dynamic_routes:
            r = dynamic_routes[key]
            name = r.gaplic_name
            roles = r.roles
            s += '  route_ref: %-12s | name:%s | roles:%s\n' % (
                key, name, roles)
        s += '\n'
        return s

    def check_routes(self):
        registry = self.registry
        static_routes = registry.static_routes
        ret = False
        for key in static_routes:
            r = static_routes[key]
            if not r.identity_ack:
                # TODO: retransmite la identity si timeout
                ret = True

            # TODO: retransmite cur_pending_event si timeout

        dynamic_routes = registry.dynamic_routes
        for key in dynamic_routes:
            r = dynamic_routes[key]
            # TODO: retransmite cur_pending_event si timeout
            pass

        return ret

    def trace_intra_event(self, prefix, intra_event):
        hora = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        message_type = intra_event.message_type

        if message_type == '__event__':
            fields = intra_event.event_fields
        elif message_type == '__event_ack__':
            fields = intra_event.event_ack_fields
        elif message_type == '__event_nack__':
            fields = intra_event.event_nack_fields
        elif message_type == '__identity_card__':
            fields = intra_event.identity_card_fields
        elif message_type == '__identity_card_ack__':
            fields = intra_event.identity_card_ack_fields
        else:
            self.logger.info(
                'ERROR unknown message type %r' % message_type)

        try:
            trace = "\n%s %s\n" % (
                hora,
                prefix,
            )
            for f in fields:
                trace += "    %s: %r\n" % (f, getattr(intra_event, f))
            self.logger.info(trace)

        except Exception as e:
            self.logger.error(
                "ERROR %s:\n%r" % (e, intra_event),
                exc_info=True,
            )

    def send_identity_card(self, route, roles_back_way=None):
        my_gaplic_name = self.registry.my_gaplic_name
        my_roles = self.registry.my_roles
        if roles_back_way:
            my_roles_back_way = roles_back_way
            prefix = '%s ^==> %s' % (my_gaplic_name, route)
        else:
            my_roles_back_way = self.registry.get_roles_back_way()
            prefix = '%s ==> %s' % (my_gaplic_name, route)

        idc = IntraEvent(
            message_type='__identity_card__',
            route_ref=None,
            serial=None,
            my_gaplic_name=my_gaplic_name,
            my_roles=list(my_roles) + list(my_roles_back_way),
            kw={},
        )

        if self.config.trace_router:
            self.trace_intra_event(prefix, idc)
        x = idc.toJSON()
        route.write(x)

    def find_local_router(self):
        if self.local_router:
            return  # already created.

        ports = list(self.config.localhost_route_ports)
        if self.server_sock:
            # remove own port if iam-server
            try:
                index = ports.index(self.server_sock.port)
                ports.pop(index)
            except ValueError:
                pass

        destinations = list((lambda x: ('localhost', x), ports))
        self.local_router = self.create_gobj(
            'local_router_client',
            GConnex,
            self,
            transmit_ready_event_name=None,
            destinations=destinations,
        )

    def start_router_server(self):
        self.server_sock = self.create_gobj(
            self.gaplic.name + '-router',
            GServerSock,
            self,
            host='0.0.0.0',
            ports=self.config.localhost_route_ports,
            use_multi_ports=True,
            transmit_ready_event_name=None,
        )

    def _send_event_to_external_route(
            self,
            route_ref,
            gaplic_name,
            role,
            gobj_name,
            event_name,
            subscriber_gobj,
            kw):
        registry = self.registry
        route = registry.get_route_by_ref(route_ref)
        if not route:
            self.logger.error("ERROR route_ref NOT FOUND: %r" % route_ref)
            return False

        intra_event = IntraEvent(
            message_type='__event__',
            serial=route.incr_serial(),
            event_name=event_name,
            destination_gaplic=gaplic_name,
            destination_role=role,
            destination_gobj=gobj_name,
            origin_gaplic=registry.my_gaplic_name,
            origin_role=registry.my_roles,  # use destination_role?
            origin_gobj=subscriber_gobj,
            kw=kw)

        registry.enqueue_pending_event(route, intra_event)
        if not route.cur_pending_event:
            registry.fire_pending_events(route)
        return True

    def make_external_subscription(self, intra_event, route_ref):
        k = random_key()
        subscription = ExternalSubscription(intra_event, route_ref)
        self.external_subscriptions[k] = subscription
        return k

    def remove_external_subscription(self, subscription_reference):
        if subscription_reference in self.external_subscriptions:
            subs = self.external_subscriptions[subscription_reference]
            if subs:
                isubscription = subs.isubscription
                named_gobj = self.gaplic.find_unique_gobj(
                    subs.destination_gobj)
                if named_gobj:
                    named_gobj.delete_subscription_by_object(isubscription)
            del self.external_subscriptions[subscription_reference]

    def get_external_subscription(self, subscription_reference):
        if subscription_reference in self.external_subscriptions:
            subs = self.external_subscriptions[subscription_reference]
            return subs
